snippets = {
	"@f": "@font-face {\n\tfont-family: ${1};\n\tsrc: url(${1});\n}",
	"@ff": "@font-face {\n\tfont-family: '${1:FontName}';\n\tsrc: url('${2:FileName}.eot');\n\tsrc: url('${2:FileName}.eot?#iefix') format('embedded-opentype'),\n\t\t url('${2:FileName}.woff') format('woff'),\n\t\t url('${2:FileName}.ttf') format('truetype'),\n\t\t url('${2:FileName}.svg#${1:FontName}') format('svg');\n\tfont-style: ${3:normal};\n\tfont-weight: ${4:normal};\n}",
	"@i|@import": "@import url(${0});",
	"@kf": "@keyframes ${1:identifier} {\n\t${2}\n}",
	"@m|@media": "@media ${1:screen} {\n\t${0}\n}",
	"ac": "align-content:start|end|flex-start|flex-end|center|space-between|space-around|stretch|space-evenly",
	"ai": "align-items:start|end|flex-start|flex-end|center|baseline|stretch",
	"anim": "animation:${1:name} ${2:duration} ${3:timing-function} ${4:delay} ${5:iteration-count} ${6:direction} ${7:fill-mode}",
	"animdel": "animation-delay:time",
	"animdir": "animation-direction:normal|reverse|alternate|alternate-reverse",
	"animdur": "animation-duration:${1:0}s",
	"animfm": "animation-fill-mode:both|forwards|backwards",
	"animic": "animation-iteration-count:1|infinite",
	"animn": "animation-name",
	"animps": "animation-play-state:running|paused",
	"animtf": "animation-timing-function:linear|ease|ease-in|ease-out|ease-in-out|cubic-bezier(${1:0.1}, ${2:0.7}, ${3:1.0}, ${3:0.1})",
	"ap": "appearance:none",
	"as": "align-self:start|end|auto|flex-start|flex-end|center|baseline|stretch",
	"b": "bottom",
	"bd": "border:${1:1px} ${2:solid} ${3:#000}",
	"bdb": "border-bottom:${1:1px} ${2:solid} ${3:#000}",
	"bdbc": "border-bottom-color:${1:#000}",
	"bdbi": "border-bottom-image:url(${0})",
	"bdbk": "border-break:close",
	"bdbli": "border-bottom-left-image:url(${0})|continue",
	"bdblrs": "border-bottom-left-radius",
	"bdbri": "border-bottom-right-image:url(${0})|continue",
	"bdbrrs": "border-bottom-right-radius",
	"bdbs": "border-bottom-style",
	"bdbw": "border-bottom-width",
	"bdc": "border-color:${1:#000}",
	"bdci": "border-corner-image:url(${0})|continue",
	"bdcl": "border-collapse:collapse|separate",
	"bdf": "border-fit:repeat|clip|scale|stretch|overwrite|overflow|space",
	"bdi": "border-image:url(${0})",
	"bdl": "border-left:${1:1px} ${2:solid} ${3:#000}",
	"bdlc": "border-left-color:${1:#000}",
	"bdlen": "border-length",
	"bdli": "border-left-image:url(${0})",
	"bdls": "border-left-style",
	"bdlw": "border-left-width",
	"bdr": "border-right:${1:1px} ${2:solid} ${3:#000}",
	"bdrc": "border-right-color:${1:#000}",
	"bdri": "border-right-image:url(${0})",
	"bdrs": "border-radius",
	"bdrst": "border-right-style",
	"bdrw": "border-right-width",
	"bds": "border-style:none|hidden|dotted|dashed|solid|double|dot-dash|dot-dot-dash|wave|groove|ridge|inset|outset",
	"bdsp": "border-spacing",
	"bdt": "border-top:${1:1px} ${2:solid} ${3:#000}",
	"bdtc": "border-top-color:${1:#000}",
	"bdti": "border-top-image:url(${0})",
	"bdtli": "border-top-left-image:url(${0})|continue",
	"bdtlrs": "border-top-left-radius",
	"bdtri": "border-top-right-image:url(${0})|continue",
	"bdtrrs": "border-top-right-radius",
	"bdts": "border-top-style",
	"bdtw": "border-top-width",
	"bdw": "border-width",
	"bfv": "backface-visibility:hidden|visible",
	"bg": "background:${1:#000}",
	"bga": "background-attachment:fixed|scroll",
	"bgbk": "background-break:bounding-box|each-box|continuous",
	"bgc": "background-color:#${1:fff}",
	"bgcp": "background-clip:padding-box|border-box|content-box|no-clip",
	"bgi": "background-image:url(${0})",
	"bgo": "background-origin:padding-box|border-box|content-box",
	"bgp": "background-position:${1:0} ${2:0}",
	"bgpx": "background-position-x",
	"bgpy": "background-position-y",
	"bgr": "background-repeat:no-repeat|repeat-x|repeat-y|space|round",
	"bgsz": "background-size:contain|cover",
	"bxsh": "box-shadow:${1:inset }${2:hoff} ${3:voff} ${4:blur} ${5:#000}|none",
	"bxsz": "box-sizing:border-box|content-box|border-box",
	"c": "color:${1:#000}",
	"cl": "clear:both|left|right|none",
	"cm": "/* ${0} */",
	"cnt": "content:'${0}'|normal|open-quote|no-open-quote|close-quote|no-close-quote|attr(${0})|counter(${0})|counters(${0})",
	"coi": "counter-increment",
	"colm": "columns",
	"colmc": "column-count",
	"colmf": "column-fill",
	"colmg": "column-gap",
	"colmr": "column-rule",
	"colmrc": "column-rule-color",
	"colmrs": "column-rule-style",
	"colmrw": "column-rule-width",
	"colms": "column-span",
	"colmw": "column-width",
	"cor": "counter-reset",
	"cp": "clip:auto|rect(${1:top} ${2:right} ${3:bottom} ${4:left})",
	"cps": "caption-side:top|bottom",
	"cur": "cursor:pointer|auto|default|crosshair|hand|help|move|pointer|text",
	"d": "display:grid|inline-grid|subgrid|block|none|flex|inline-flex|inline|inline-block|list-item|run-in|compact|table|inline-table|table-caption|table-column|table-column-group|table-header-group|table-footer-group|table-row|table-row-group|table-cell|ruby|ruby-base|ruby-base-group|ruby-text|ruby-text-group",
	"ec": "empty-cells:show|hide",
	"f": "font:${1:1em} ${2:sans-serif}",
	"fd": "font-display:auto|block|swap|fallback|optional",
	"fef": "font-effect:none|engrave|emboss|outline",
	"fem": "font-emphasize",
	"femp": "font-emphasize-position:before|after",
	"fems": "font-emphasize-style:none|accent|dot|circle|disc",
	"ff": "font-family:serif|sans-serif|cursive|fantasy|monospace",
	"fft": "font-family:\"Times New Roman\", Times, Baskerville, Georgia, serif",
	"ffa": "font-family:Arial, \"Helvetica Neue\", Helvetica, sans-serif",
	"ffv": "font-family:Verdana, Geneva, sans-serif",
	"fl": "float:left|right|none",
	"fs": "font-style:italic|normal|oblique",
	"fsm": "font-smoothing:antialiased|subpixel-antialiased|none",
	"fst": "font-stretch:normal|ultra-condensed|extra-condensed|condensed|semi-condensed|semi-expanded|expanded|extra-expanded|ultra-expanded",
	"fv": "font-variant:normal|small-caps",
	"fvs": "font-variation-settings:normal|inherit|initial|unset",
	"fw": "font-weight:normal|bold|bolder|lighter",
	"fx": "flex",
	"fxb": "flex-basis:fill|max-content|min-content|fit-content|content",
	"fxd": "flex-direction:row|row-reverse|column|column-reverse",
	"fxf": "flex-flow",
	"fxg": "flex-grow",
	"fxsh": "flex-shrink",
	"fxw": "flex-wrap:nowrap|wrap|wrap-reverse",
	"fsz": "font-size",
	"fsza": "font-size-adjust",
	"gtc": "grid-template-columns:repeat()|minmax()",
	"gtr": "grid-template-rows:repeat()|minmax()",
	"gta": "grid-template-areas",
	"gt": "grid-template",
	"gg": "grid-gap",
	"gcg": "grid-column-gap",
	"grg": "grid-row-gap",
	"gac": "grid-auto-columns:auto|minmax()",
	"gar": "grid-auto-rows:auto|minmax()",
	"gaf": "grid-auto-flow:row|column|dense|inherit|initial|unset",
	"gd": "grid",
	"gc": "grid-column",
	"gcs": "grid-column-start",
	"gce": "grid-column-end",
	"gr": "grid-row",
	"grs": "grid-row-start",
	"gre": "grid-row-end",
	"ga": "grid-area",
	"h": "height",
	"jc": "justify-content:start|end|stretch|flex-start|flex-end|center|space-between|space-around|space-evenly",
	"ji": "justify-items:start|end|center|stretch",
	"js": "justify-self:start|end|center|stretch",
	"l": "left",
	"lg": "background-image:linear-gradient(${1})",
	"lh": "line-height",
	"lis": "list-style",
	"lisi": "list-style-image",
	"lisp": "list-style-position:inside|outside",
	"list": "list-style-type:disc|circle|square|decimal|decimal-leading-zero|lower-roman|upper-roman",
	"lts": "letter-spacing:normal",
	"m": "margin",
	"mah": "max-height",
	"mar": "max-resolution",
	"maw": "max-width",
	"mb": "margin-bottom",
	"mih": "min-height",
	"mir": "min-resolution",
	"miw": "min-width",
	"ml": "margin-left",
	"mr": "margin-right",
	"mt": "margin-top",
	"ol": "outline",
	"olc": "outline-color:${1:#000}|invert",
	"olo": "outline-offset",
	"ols": "outline-style:none|dotted|dashed|solid|double|groove|ridge|inset|outset",
	"olw": "outline-width|thin|medium|thick",
	"op": "opacity",
	"ord": "order",
	"ori": "orientation:landscape|portrait",
	"orp": "orphans",
	"ov": "overflow:hidden|visible|hidden|scroll|auto",
	"ovs": "overflow-style:scrollbar|auto|scrollbar|panner|move|marquee",
	"ovx": "overflow-x:hidden|visible|hidden|scroll|auto",
	"ovy": "overflow-y:hidden|visible|hidden|scroll|auto",
	"p": "padding",
	"pb": "padding-bottom",
	"pgba": "page-break-after:auto|always|left|right",
	"pgbb": "page-break-before:auto|always|left|right",
	"pgbi": "page-break-inside:auto|avoid",
	"pl": "padding-left",
	"pos": "position:relative|absolute|relative|fixed|static",
	"pr": "padding-right",
	"pt": "padding-top",
	"q": "quotes",
	"qen": "quotes:'\\201C' '\\201D' '\\2018' '\\2019'",
	"qru": "quotes:'\\00AB' '\\00BB' '\\201E' '\\201C'",
	"r": "right",
	"rsz": "resize:none|both|horizontal|vertical",
	"t": "top",
	"ta": "text-align:left|center|right|justify",
	"tal": "text-align-last:left|center|right",
	"tbl": "table-layout:fixed",
	"td": "text-decoration:none|underline|overline|line-through",
	"te": "text-emphasis:none|accent|dot|circle|disc|before|after",
	"th": "text-height:auto|font-size|text-size|max-size",
	"ti": "text-indent",
	"tj": "text-justify:auto|inter-word|inter-ideograph|inter-cluster|distribute|kashida|tibetan",
	"to": "text-outline:${1:0} ${2:0} ${3:#000}",
	"tov": "text-overflow:ellipsis|clip",
	"tr": "text-replace",
	"trf": "transform:${1}|skewX(${1:angle})|skewY(${1:angle})|scale(${1:x}, ${2:y})|scaleX(${1:x})|scaleY(${1:y})|scaleZ(${1:z})|scale3d(${1:x}, ${2:y}, ${3:z})|rotate(${1:angle})|rotateX(${1:angle})|rotateY(${1:angle})|rotateZ(${1:angle})|translate(${1:x}, ${2:y})|translateX(${1:x})|translateY(${1:y})|translateZ(${1:z})|translate3d(${1:tx}, ${2:ty}, ${3:tz})",
	"trfo": "transform-origin",
	"trfs": "transform-style:preserve-3d",
	"trs": "transition:${1:prop} ${2:time}",
	"trsde": "transition-delay:${1:time}",
	"trsdu": "transition-duration:${1:time}",
	"trsp": "transition-property:${1:prop}",
	"trstf": "transition-timing-function:${1:fn}",
	"tsh": "text-shadow:${1:hoff} ${2:voff} ${3:blur} ${4:#000}",
	"tt": "text-transform:uppercase|lowercase|capitalize|none",
	"tw": "text-wrap:none|normal|unrestricted|suppress",
	"us": "user-select:none",
	"v": "visibility:hidden|visible|collapse",
	"va": "vertical-align:top|super|text-top|middle|baseline|bottom|text-bottom|sub",
	"w": "width",
	"whs": "white-space:nowrap|pre|pre-wrap|pre-line|normal",
	"whsc": "white-space-collapse:normal|keep-all|loose|break-strict|break-all",
	"wid": "widows",
	"wm": "writing-mode:lr-tb|lr-tb|lr-bt|rl-tb|rl-bt|tb-rl|tb-lr|bt-lr|bt-rl",
	"wob": "word-break:normal|keep-all|break-all",
	"wos": "word-spacing",
	"wow": "word-wrap:none|unrestricted|suppress|break-word|normal",
	"z": "z-index",
	"zom": "zoom:1"
}



from .scan import scan, TokenType
from ..scanner_utils import is_space
from .parse import split_value

class MatchResult:
    __slots__ = ('type', 'start', 'end', 'body_start', 'body_end')

    def __init__(self, match_type: str, start: int, end: int, body_start: int=None, body_end: int=None):
        self.type = match_type
        self.start = start
        self.end = end
        self.body_start = body_start
        self.body_end = body_end

    def to_json(self):
        return {
            'type': self.type,
            'start': self.start,
            'end': self.end,
            'body_start': self.body_start,
            'body_end': self.body_end
        }

class InwardRange:
    __slots__ = ('start', 'end', 'delimiter', 'first_child')

    def __init__(self, start: int, end: int, delimiter: int):
        self.start = start
        self.end = end
        self.delimiter = delimiter
        self.first_child = None


def match(source: str, pos: int) -> MatchResult:
    pool = []
    stack = []
    result = [None]
    pending_property = []
    pending_property.append(None) # get rid of pyLint rant

    def release_pending():
        if pending_property[0]:
            release_range(pool, pending_property[0])
            pending_property[0] = None

    def scan_callback(token_type: str, start: int, end: int, delimiter: int):
        if token_type == TokenType.Selector:
            release_pending()
            stack.append(alloc_range(pool, start, end, delimiter))
        elif token_type == TokenType.BlockEnd:
            release_pending()
            parent = stack and stack.pop()
            if parent and parent[0] < pos < end:
                result[0] = MatchResult('selector', parent[0], end, parent[2] + 1, start)
                return False
        elif token_type == TokenType.PropertyName:
            release_pending()
            pending_property[0] = alloc_range(pool, start, end, delimiter)
        elif token_type == TokenType.PropertyValue:
            pending = pending_property[0]
            if pending and pending[0] < pos < end:
                result[0] = MatchResult('property', pending[0], delimiter + 1, start, end)
                return False
            release_pending()


    scan(source, scan_callback)
    return result[0]


def balanced_outward(source: str, pos: int) -> list:
    """
    Returns balanced CSS model: a list of all ranges that could possibly match
    given location when moving in outward direction
    """
    pool = []
    stack = []
    result = []
    prop = []
    prop.append(None) # Get rid of pyLint rant

    def scan_callback(token_type: str, start: int, end: int, delimiter: int):
        if token_type == TokenType.Selector:
            stack.append(alloc_range(pool, start, end, delimiter))
        elif token_type == TokenType.BlockEnd:
            left = stack and stack.pop()
            if left and left[0] < pos < end:
                # Matching section found
                inner = inner_range(source, left[2] + 1, start)
                if inner:
                    push(result, inner)
                push(result, (left[0], end))
            if left:
                release_range(pool, left)
            if not stack:
                return False
        elif token_type == TokenType.PropertyName:
            if prop[0]:
                release_range(pool, prop[0])

            prop[0] = alloc_range(pool, start, end, delimiter)
        elif token_type == TokenType.PropertyValue:
            p = prop[0]
            if p and p[0] < pos < max(delimiter, end):
                # Push full token and value range
                push(result, (start, end))
                push(result, (p[0], delimiter + 1 if delimiter != -1 else end))

        if token_type != TokenType.PropertyName and prop[0]:
            release_range(pool, prop[0])
            prop[0] = None

    scan(source, scan_callback)
    return result


def balanced_inward(source: str, pos: int) -> list:
    """
    Returns balanced CSS selectors: a list of all ranges that could possibly match
    given location when moving in inward direction
    """
    # Collecting ranges for inward balancing is a bit trickier: we have to store
    # first child of every matched selector until we find the one that matches given
    # location
    pool = []
    stack = []
    result = []
    pending_property = []
    pending_property.append(None)

    def alloc(start: int, end: int, delimiter: int):
        if pool:
            r = pool.pop()
            r.start = start
            r.end = end
            r.delimiter = delimiter
            return r

        return InwardRange(start, end, delimiter)

    def release(r: InwardRange):
        r.first_child = None
        pool.append(r)

    def release_pending():
        if pending_property[0]:
            release(pending_property[0])
            pending_property[0] = None

    def push_child(start: int, end: int, delimiter: int):
        """
        Pushes given inward range as a first child of current selector only if it’s
        not set yet
        """
        parent = stack and stack[-1]
        if parent and not parent.first_child:
            parent.first_child = alloc(start, end, delimiter)

    def scan_callback(token_type: str, start: int, end: int, delimiter: int):
        if token_type == TokenType.BlockEnd:
            release_pending()

            r = stack and stack.pop()
            if not r:
                # Some sort of lone closing brace, ignore it
                return

            if r.start <= pos <= end:
                # Matching selector found: add it and its inner range into result
                inner = inner_range(source, r.delimiter + 1, start)
                push(result, (r.start, end))
                if inner:
                    push(result, inner)

                while r.first_child:
                    child = r.first_child

                    inner = inner_range(source, child.delimiter + 1, child.end - 1)
                    push(result, (child.start, child.end))
                    if inner:
                        push(result, inner)
                    r = child

                return False
            else:
                parent = stack and stack[-1]
                if parent and not parent.first_child:
                    # No first child in parent node: store current selector
                    r.end = end
                    parent.first_child = r
                else:
                    release(r)
        elif token_type == TokenType.PropertyName:
            release_pending()
            pending_property[0] = alloc(start, end, delimiter)
            push_child(start, end, delimiter)
        elif token_type == TokenType.PropertyValue:
            if pending_property[0]:
                p = pending_property[0]
                if p.start <= pos <= end:
                    # Direct hit into property, no need to look further
                    push(result, (p.start, delimiter + 1))
                    push(result, (start, end))
                    release_pending()
                    return False

                parent = stack and stack[-1]
                if parent and parent.first_child and parent.first_child.start == p.start:
                    # First child is an expected property name, update its range
                    # to include property value
                    parent.first_child.end = delimiter + 1 if delimiter != -1 else end

                release_pending()
        else:
            # Selector start
            stack.append(alloc(start, end, delimiter))
            release_pending()

    scan(source, scan_callback)
    return result


def inner_range(source: str, start: int, end: int):
    """
    Returns inner range for given selector bounds: narrows it to first non-empty
    region. If resulting region is empty, returns `None`
    """
    while start < end and is_space(source[start]):
        start += 1

    while end and end > start and is_space(source[end - 1]):
        end -= 1

    return (start, end) if start != end else None

def alloc_range(pool: list, start: int, end: int, delimiter: int):
    if pool:
        r = pool.pop()
        r[0] = start
        r[1] = end
        r[2] = delimiter
        return r

    return [start, end, delimiter]

def release_range(pool: list, r: list):
    if r:
        pool.append(r)


def push(ranges: list, r: list):
    prev = ranges and ranges[-1]
    if (not prev or prev[0] != r[0] or prev[1] != r[1]) and r[0] != r[1]:
        ranges.append(r)



from ..abbreviation import AbbreviationNode
from ..config import Config
from ..output_stream import is_inline

ELEMENT_MAP = {
    'p': 'span',
    'ul': 'li',
    'ol': 'li',
    'table': 'tr',
    'tr': 'td',
    'tbody': 'tr',
    'thead': 'tr',
    'tfoot': 'tr',
    'colgroup': 'col',
    'select': 'option',
    'optgroup': 'option',
    'audio': 'source',
    'video': 'source',
    'object': 'param',
    'map': 'area'
}

def implicit_tag(node: AbbreviationNode, ancestors: list, config: Config):
    if not node.name and node.attributes:
        resolve_implicit_tag(node, ancestors, config)


def resolve_implicit_tag(node: AbbreviationNode, ancestors: list, config: Config):
    global ELEMENT_MAP
    parent = get_parent_element(ancestors)
    context_name = config.get('context').get('name', '') if config.get('context') else ''
    parent_name = lowercase(parent.name if parent else context_name)
    node.name = ELEMENT_MAP.get(parent_name, 'span' if is_inline(parent_name, config) else 'div')


def lowercase(text: str=None):
    return (text or '').lower()

def get_parent_element(ancestors: list):
    "Returns closest element node from given ancestors list"
    i = len(ancestors) - 1
    while i >= 0:
        elem = ancestors[i]
        if isinstance(elem, AbbreviationNode):
            return elem
        i -= 1



from .walk import WalkState
from .utils import push_tokens
from .template import template
from ...abbreviation import AbbreviationNode
from ...config import Config


class CommentWalkState(WalkState):
    __slots__ = ('enabled', 'trigger', 'before', 'after')

    def __init__(self, config: Config):
        super(CommentWalkState, self).__init__(config)
        options = config.options
        before = options.get('comment.before')
        after = options.get('comment.after')

        self.enabled = options.get('comment.enabled')
        self.trigger = options.get('comment.trigger')
        self.before = template(before) if before else None
        self.after = template(after) if after else None


def comment_node_before(node: AbbreviationNode, state):
    "Adds comment prefix for given node, if required"
    if should_comment(node, state) and state.comment.before:
        output(node, state.comment.before, state)


def comment_node_after(node: AbbreviationNode, state):
    "Adds comment suffix for given node, if required"
    if should_comment(node, state) and state.comment.after:
        output(node, state.comment.after, state)


def should_comment(node: AbbreviationNode, state):
    "Check if given node should be commented"
    comment = state.comment

    if not comment.enabled or not comment.trigger or not node.name or not node.attributes:
        return False

    for attr in node.attributes:
        if attr.name and attr.name in comment.trigger:
            return True

    return False


def output(node: AbbreviationNode, tokens: list, state: WalkState):
    "Pushes given template tokens into output stream"
    attrs = {}
    out = state.out

    # Collect attributes payload
    for attr in node.attributes:
        if attr.name and attr.value:
            attrs[attr.name.upper()] = attr.value

    # Output parsed tokens
    for token in tokens:
        if isinstance(token, str):
            out.push_string(token)
        elif token.name in attrs:
            out.push_string(token.before)
            push_tokens(attrs[token.name], state)
            out.push_string(token.after)



class Token:
    __slots__ = ('start', 'end')

    def __init__(self, start: int=None, end: int=None):
        self.start = start
        self.end = end

    @property
    def type(self):
        "Type of current token"
        return self.__class__.__name__

    def to_json(self):
        return dict([(k, self.__getattribute__(k)) for k in dir(self) if not k.startswith('__') and k != 'to_json'])

class Repeater(Token):
    __slots__ = ('count', 'value', 'implicit')

    def __init__(self, count: int, value: int, implicit: bool=False, *args):
        super(Repeater, self).__init__(*args)
        self.count = count
        self.value = value
        self.implicit = implicit

class RepeaterNumber(Token):
    __slots__ = ('size', 'reverse', 'base', 'parent')

    def __init__(self, size: int, reverse: bool, base: int=0, parent: int=0, *args):
        super(RepeaterNumber, self).__init__(*args)
        self.size = size
        self.reverse = reverse
        self.base = base
        self.parent = parent


class RepeaterPlaceholder(Token):
    __slots__ = ('value',)

    def __init__(self, value: str=None, *args):
        super(RepeaterPlaceholder, self).__init__(*args)
        self.value = value

class Field(Token):
    __slots__ = ('name', 'index')

    def __init__(self, name: str, index: int=None, *args):
        super(Field, self).__init__(*args)
        self.index = index
        self.name = name

class Operator(Token):
    __slots__ = ('operator',)

    def __init__(self, operator: str, *args):
        super(Operator, self).__init__(*args)
        self.operator = operator

class Bracket(Token):
    __slots__ = ('open', 'context')

    def __init__(self, is_open: bool, context: str, *args):
        super(Bracket, self).__init__(*args)
        self.open = is_open
        self.context = context

class Quote(Token):
    __slots__ = ('single', )

    def __init__(self, single: bool, *args):
        super(Quote, self).__init__(*args)
        self.single = single

class Literal(Token):
    __slots__ = ('value',)

    def __init__(self, value: str, *args):
        super(Literal, self).__init__(*args)
        self.value = value

class WhiteSpace(Token): pass



from .tokenizer import tokenize, tokens
from .parser import parser, CSSProperty, CSSValue, FunctionCall
from ..scanner import ScannerException

def parse(abbr: str, options={}):
    "Parses given abbreviation into property set"
    try:
        tokens = tokenize(abbr, options.get('value', False)) if isinstance(abbr, str) else abbr
        return parser(tokens, options)
    except ScannerException as err:
        if isinstance(abbr, str):
            err.message += '\n%s\n%s^' % (abbr, '-' * err.pos)

        raise err



from ...abbreviation import AbbreviationNode, AbbreviationAttribute

def jsx(node: AbbreviationNode):
    """
    JSX transformer: replaces `class` and `for` attributes with `className` and
    `htmlFor` attributes respectively
    """
    if node.attributes:
        for attr in node.attributes:
            rename(attr)

def rename(attr: AbbreviationAttribute):
    if attr.name == 'class':
        attr.name = 'className'
    elif attr.name == 'for':
        attr.name = 'htmlFor'



from ...token_scanner import TokenScanner, TokenScannerException
from ..tokenizer import tokens

class TokenAttribute:
    __slots__ = ('name', 'value', 'expression')

    def __init__(self, name: list=None, value: list=None, expression: bool=False):
        self.name = name
        self.value = value
        self.expression = expression


class TokenElement:
    __slots__ = ('type', 'name', 'attributes', 'value', 'repeat', 'self_close', 'elements')

    def __init__(self):
        self.type = 'TokenElement'
        self.name = None
        self.attributes = None
        self.value = None
        self.repeat = None
        self.self_close = False
        self.elements = []


class TokenGroup:
    __slots__ = ('type', 'elements', 'repeat')

    def __init__(self):
        self.type = 'TokenGroup'
        self.elements = []
        self.repeat = None


def parse(abbr: list, options: dict={}):
    scanner = TokenScanner(abbr)
    result = statements(scanner, options)
    if scanner.readable():
        raise scanner.error('Unexpected character')

    return result


def statements(scanner: TokenScanner, options: dict):
    result = TokenGroup()
    ctx = result
    node = None
    stack = []

    while scanner.readable():
        node = element(scanner, options) or group(scanner, options)
        if node:
            ctx.elements.append(node)
            if scanner.consume(is_child_operator):
                stack.append(ctx)
                ctx = node
            elif scanner.consume(is_sibling_operator):
                continue
            elif is_climb_operator(scanner.peek()):
                while scanner.consume(is_climb_operator):
                    if len(stack): ctx = stack.pop()
        else:
            break

    return result


def group(scanner: TokenScanner, options: dict):
    "Consumes group from given scanner"
    if scanner.consume(is_group_start):
        result = statements(scanner, options)
        token = scanner.next()
        if is_bracket(token, 'group', False):
            result.repeat = repeater(scanner)
        return result


def element(scanner: TokenScanner, options: dict):
    "Consumes single element from given scanner"
    attr = None
    elem = TokenElement()

    if element_name(scanner, options):
        elem.name = scanner.slice()

    while scanner.readable():
        scanner.start = scanner.pos
        if not elem.repeat and not is_empty(elem) and scanner.consume(is_repeater):
            elem.repeat = scanner.tokens[scanner.pos - 1]
        elif elem.value is None and text(scanner):
            elem.value = get_text(scanner)
        else:
            attr = short_attribute(scanner, 'id', options) or short_attribute(scanner, 'class', options) or attribute_set(scanner)
            if attr is not None:
                if not isinstance(attr, list):
                    attr = [attr]
                if elem.attributes is None:
                    elem.attributes = attr[:]
                else:
                    elem.attributes += attr
            else:
                if not is_empty(elem) and scanner.consume(is_close_operator):
                    elem.self_close = True
                    if not elem.repeat and scanner.consume(is_repeater):
                        elem.repeat = scanner.tokens[scanner.pos - 1]
                break

    return elem if not is_empty(elem) else None


def attribute_set(scanner: TokenScanner):
    "Consumes attribute set from given scanner"
    if scanner.consume(is_attribute_set_start):
        attributes = []
        attr = None

        while scanner.readable():
            attr = attribute(scanner)
            if attr:
                attributes.append(attr)
            elif scanner.consume(is_attribute_set_end):
                break
            elif not scanner.consume(is_white_space):
                raise scanner.error('Unexpected "%s" token' % scanner.peek().type)

        return attributes


def short_attribute(scanner: TokenScanner, attr_type: str, options: dict):
    "Consumes attribute shorthand (class or id) from given scanner"
    if is_operator(scanner.peek(), attr_type):
        scanner.pos += 1
        attr = TokenAttribute([create_literal(attr_type)])

        # Consume expression after shorthand start for React-like components
        if options.get('jsx') and text(scanner):
            attr.value = get_text(scanner)
            attr.expression = True
        else:
            attr.value = scanner.slice() if literal(scanner) else None

        return attr


def attribute(scanner: TokenScanner):
    if quoted(scanner):
        # Consumed quoted value: it’s a value for default attribute
        return TokenAttribute(value=scanner.slice())

    if literal(scanner, True):
        name = scanner.slice()
        value = None
        if scanner.consume(is_equals) and (quoted(scanner) or literal(scanner, True)):
            value = scanner.slice()
        return TokenAttribute(name, value)


def repeater(scanner: TokenScanner):
    if is_repeater(scanner.peek()):
        return scanner.next()


def quoted(scanner: TokenScanner):
    "Consumes quoted value from given scanner, if possible"
    start = scanner.pos
    quote = scanner.peek()
    if is_quote(quote):
        scanner.pos += 1
        while scanner.readable():
            if is_quote(scanner.next(), quote.single):
                scanner.start = start
                return True

        raise scanner.error('Unclosed quote', quote)

    return False


def literal(scanner: TokenScanner, allow_brackets=False):
    "Consumes literal (unquoted value) from given scanner"
    start = scanner.pos
    brackets = {
        'attribute': 0,
        'expression': 0,
        'group': 0
    }

    while scanner.readable():
        token = scanner.peek()
        if brackets['expression']:
            # If we’re inside expression, we should consume all content in it
            if is_bracket(token, 'expression'):
                brackets[token.context] += 1 if token.open else -1
        elif is_quote(token) or is_operator(token) or is_white_space(token) or is_repeater(token):
            break
        elif is_bracket(token):
            if not allow_brackets: break

            if token.open:
                brackets[token.context] += 1
            elif not brackets[token.context]:
                # Stop if found unmatched closing brace: it must be handled
                # by parent consumer
                break
            else:
                brackets[token.context] -= 1

        scanner.pos += 1

    if start != scanner.pos:
        scanner.start = start
        return True

    return False


def element_name(scanner: TokenScanner, options: dict):
    "Consumes element name from given scanner"
    start = scanner.pos

    if options.get('jsx') and scanner.consume(is_capitalized_literal):
        # Check for edge case: consume immediate capitalized class names
        # for React-like components, e.g. `Foo.Bar.Baz`
        while scanner.readable():
            pos = scanner.pos
            if not scanner.consume(is_class_name_operator) or not scanner.consume(is_capitalized_literal):
                scanner.pos = pos
                break

    while scanner.readable() and scanner.consume(is_element_name):
        pass

    if scanner.pos != start:
        scanner.start = start
        return True

    return False


def text(scanner: TokenScanner):
    "Consumes text value from given scanner"
    start = scanner.pos
    if scanner.consume(is_text_start):
        brackets = 0
        while scanner.readable():
            token = scanner.next()
            if is_bracket(token, 'expression'):
                if token.open:
                    brackets += 1
                elif not brackets:
                    break
                else:
                    brackets -= 1

        scanner.start = start
        return True

    return False


def get_text(scanner: TokenScanner):
    start = scanner.start
    end = scanner.pos
    if is_bracket(scanner.tokens[start], 'expression', True):
        start += 1

    if is_bracket(scanner.tokens[end - 1], 'expression', False):
        end -= 1

    return scanner.slice(start, end)


def is_bracket(token: tokens.Bracket, context: str=None, is_open=None):
    return isinstance(token, tokens.Bracket) and \
        (context is None or token.context == context) and \
        (is_open is None or token.open == is_open)


def is_operator(token: tokens.Operator, op_type: str=None):
    return isinstance(token, tokens.Operator) and (not op_type or token.operator == op_type)


def is_quote(token: tokens.Quote, is_single=None):
    return isinstance(token, tokens.Quote) and (is_single is None or token.single == is_single)


def is_white_space(token: tokens.WhiteSpace):
    return isinstance(token, tokens.WhiteSpace)


def is_equals(token: tokens.Operator):
    return is_operator(token, 'equal')


def is_repeater(token: tokens.Repeater):
    return isinstance(token, tokens.Repeater)


def is_literal(token: tokens.Literal):
    return isinstance(token, tokens.Literal)


def is_capitalized_literal(token: tokens.Literal):
    if is_literal(token):
        return 'A' <= token.value[0] <= 'Z'

    return False


def is_element_name(token: tokens.Literal):
    return is_literal(token) or isinstance(token, tokens.RepeaterNumber) or isinstance(token, tokens.RepeaterPlaceholder)


def is_class_name_operator(token: tokens.Operator):
    return is_operator(token, 'class')


def is_attribute_set_start(token: tokens.Bracket):
    return is_bracket(token, 'attribute', True)


def is_attribute_set_end(token: tokens.Bracket):
    return is_bracket(token, 'attribute', False)


def is_text_start(token: tokens.Bracket):
    return is_bracket(token, 'expression', True)


def is_group_start(token: tokens.Bracket):
    return is_bracket(token, 'group', True)


def create_literal(value: str):
    return tokens.Literal(value)


def is_empty(elem: TokenElement):
    return elem.name is None and elem.value is None and elem.attributes is None


def is_child_operator(token: tokens.Operator):
    return is_operator(token, 'child')


def is_sibling_operator(token: tokens.Operator):
    return is_operator(token, 'sibling')


def is_climb_operator(token: tokens.Operator):
    return is_operator(token, 'climb')


def is_close_operator(token: tokens.Operator):
    return is_operator(token, 'close')



class Brackets:
    SquareL = '['
    SquareR = ']'
    RoundL = '('
    RoundR = ')'
    CurlyL = '{'
    CurlyR = '}'

BRACE_PAIRS = dict([
    (Brackets.SquareL, Brackets.SquareR),
    (Brackets.RoundL, Brackets.RoundR),
    (Brackets.CurlyL, Brackets.CurlyR),
])



from ..config import Config
from ..abbreviation import AbbreviationNode, AbbreviationAttribute

def merge_attributes(node: AbbreviationNode, config: Config):
    """
    Merges attributes in current node: de-duplicates attributes with the same name
    and merges class names
    """
    if not node.attributes:
        return

    attributes = []
    lookup = {}

    for attr in node.attributes:
        if attr.name:
            attr_name = attr.name
            if attr_name in lookup:
                prev = lookup[attr_name]
                if attr_name == 'class':
                    prev.value = merge_value(prev.value, attr.value, ' ')
                else:
                    merge_declarations(prev, attr, config)
            else:
                # Create new attribute instance so we can safely modify it later
                lookup[attr_name] = attr.copy()
                attributes.append(lookup[attr_name])
        else:
            attributes.append(attr)

    node.attributes = attributes

def merge_value(prev_value: list=None, next_value: list=None, glue: str=''):
    "Merges two token lists into single list. Adjacent strings are merged together"
    if prev_value is not None and next_value is not None:
        if prev_value and glue:
            append(prev_value, glue)

        for t in next_value:
            append(prev_value, t)

        return prev_value

    result = prev_value or next_value
    return result and result[:]

def merge_declarations(dest: AbbreviationAttribute, src: AbbreviationAttribute, config: Config):
    "Merges data from `src` attribute into `dest` and returns it"
    dest.name = src.name

    if not config.options.get('output.reverseAttributes'):
        dest.value = src.value

    # Keep high-priority properties
    if not dest.implied: dest.implied = src.implied
    if not dest.boolean: dest.boolean = src.boolean

    if dest.value_type != 'expression':
        dest.value_type = src.value_type

    return dest

def append(tokens: list, value: str):
    if tokens and isinstance(tokens[-1], str) and isinstance(value, str):
        tokens[-1] += value
    else:
        tokens.append(value)



snippets = {
	"a": "a[href]",
	"a:blank": "a[href='http://${0}' target='_blank' rel='noopener noreferrer']",
	"a:link": "a[href='http://${0}']",
	"a:mail": "a[href='mailto:${0}']",
	"a:tel": "a[href='tel:+${0}']",
	"abbr": "abbr[title]",
	"acr|acronym": "acronym[title]",
	"base": "base[href]/",
	"basefont": "basefont/",
	"br": "br/",
	"frame": "frame/",
	"hr": "hr/",
	"bdo": "bdo[dir]",
	"bdo:r": "bdo[dir=rtl]",
	"bdo:l": "bdo[dir=ltr]",
	"col": "col/",
	"link": "link[rel=stylesheet href]/",
	"link:css": "link[href='${1:style}.css']",
	"link:print": "link[href='${1:print}.css' media=print]",
	"link:favicon": "link[rel='shortcut icon' type=image/x-icon href='${1:favicon.ico}']",
	"link:mf|link:manifest": "link[rel='manifest' href='${1:manifest.json}']",
	"link:touch": "link[rel=apple-touch-icon href='${1:favicon.png}']",
	"link:rss": "link[rel=alternate type=application/rss+xml title=RSS href='${1:rss.xml}']",
	"link:atom": "link[rel=alternate type=application/atom+xml title=Atom href='${1:atom.xml}']",
	"link:im|link:import": "link[rel=import href='${1:component}.html']",
	"meta": "meta/",
	"meta:utf": "meta[http-equiv=Content-Type content='text/html;charset=UTF-8']",
	"meta:vp": "meta[name=viewport content='width=${1:device-width}, initial-scale=${2:1.0}']",
	"meta:compat": "meta[http-equiv=X-UA-Compatible content='${1:IE=7}']",
	"meta:edge": "meta:compat[content='${1:ie=edge}']",
	"meta:redirect": "meta[http-equiv=refresh content='0; url=${1:http://example.com}']",
	"meta:kw": "meta[name=keywords content]",
	"meta:desc": "meta[name=description content]",
	"style": "style",
	"script": "script",
	"script:src": "script[src]",
	"img": "img[src alt]/",
	"img:s|img:srcset": "img[srcset src alt]",
	"img:z|img:sizes": "img[sizes srcset src alt]",
	"picture": "picture",
	"src|source": "source/",
	"src:sc|source:src": "source[src type]",
	"src:s|source:srcset": "source[srcset]",
	"src:t|source:type": "source[srcset type='${1:image/}']",
	"src:z|source:sizes": "source[sizes srcset]",
	"src:m|source:media": "source[media='(${1:min-width: })' srcset]",
	"src:mt|source:media:type": "source:media[type='${2:image/}']",
	"src:mz|source:media:sizes": "source:media[sizes srcset]",
	"src:zt|source:sizes:type": "source[sizes srcset type='${1:image/}']",
	"iframe": "iframe[src frameborder=0]",
	"embed": "embed[src type]/",
	"object": "object[data type]",
	"param": "param[name value]/",
	"map": "map[name]",
	"area": "area[shape coords href alt]/",
	"area:d": "area[shape=default]",
	"area:c": "area[shape=circle]",
	"area:r": "area[shape=rect]",
	"area:p": "area[shape=poly]",
	"form": "form[action]",
	"form:get": "form[method=get]",
	"form:post": "form[method=post]",
	"label": "label[for]",
	"input": "input[type=${1:text}]/",
	"inp": "input[name=${1} id=${1}]",
	"input:h|input:hidden": "input[type=hidden name]",
	"input:t|input:text": "inp[type=text]",
	"input:search": "inp[type=search]",
	"input:email": "inp[type=email]",
	"input:url": "inp[type=url]",
	"input:p|input:password": "inp[type=password]",
	"input:datetime": "inp[type=datetime]",
	"input:date": "inp[type=date]",
	"input:datetime-local": "inp[type=datetime-local]",
	"input:month": "inp[type=month]",
	"input:week": "inp[type=week]",
	"input:time": "inp[type=time]",
	"input:tel": "inp[type=tel]",
	"input:number": "inp[type=number]",
	"input:color": "inp[type=color]",
	"input:c|input:checkbox": "inp[type=checkbox]",
	"input:r|input:radio": "inp[type=radio]",
	"input:range": "inp[type=range]",
	"input:f|input:file": "inp[type=file]",
	"input:s|input:submit": "input[type=submit value]",
	"input:i|input:image": "input[type=image src alt]",
	"input:b|input:button": "input[type=button value]",
	"input:reset": "input:button[type=reset]",
	"isindex": "isindex/",
	"select": "select[name=${1} id=${1}]",
	"select:d|select:disabled": "select[disabled.]",
	"opt|option": "option[value]",
	"textarea": "textarea[name=${1} id=${1} cols=${2:30} rows=${3:10}]",
	"marquee": "marquee[behavior direction]",
	"menu:c|menu:context": "menu[type=context]",
	"menu:t|menu:toolbar": "menu[type=toolbar]",
	"video": "video[src]",
	"audio": "audio[src]",
	"html:xml": "html[xmlns=http://www.w3.org/1999/xhtml]",
	"keygen": "keygen/",
	"command": "command/",
	"btn:s|button:s|button:submit" : "button[type=submit]",
	"btn:r|button:r|button:reset" : "button[type=reset]",
	"btn:d|button:d|button:disabled" : "button[disabled.]",
	"fst:d|fset:d|fieldset:d|fieldset:disabled" : "fieldset[disabled.]",

	"bq": "blockquote",
	"fig": "figure",
	"figc": "figcaption",
	"pic": "picture",
	"ifr": "iframe",
	"emb": "embed",
	"obj": "object",
	"cap": "caption",
	"colg": "colgroup",
	"fst": "fieldset",
	"btn": "button",
	"optg": "optgroup",
	"tarea": "textarea",
	"leg": "legend",
	"sect": "section",
	"art": "article",
	"hdr": "header",
	"ftr": "footer",
	"adr": "address",
	"dlg": "dialog",
	"str": "strong",
	"prog": "progress",
	"mn": "main",
	"tem": "template",
	"fset": "fieldset",
	"datag": "datagrid",
	"datal": "datalist",
	"kg": "keygen",
	"out": "output",
	"det": "details",
	"sum": "summary",
	"cmd": "command",

	"ri:d|ri:dpr": "img:s",
	"ri:v|ri:viewport": "img:z",
	"ri:a|ri:art": "pic>src:m+img",
	"ri:t|ri:type": "pic>src:t+img",

	"!!!": "{<!DOCTYPE html>}",
	"doc": "html[lang=${lang}]>(head>meta[charset=${charset}]+meta:vp+title{${1:Document}})+body",
	"!|html:5": "!!!+doc",

	"c": "{<!-- ${0} -->}",
	"cc:ie": "{<!--[if IE]>${0}<![endif]-->}",
	"cc:noie": "{<!--[if !IE]><!-->${0}<!--<![endif]-->}"
}



def calculate_score(str1: str, str2: str, partial_match=False):
    """
    Calculates how close `str1` matches `str2` using fuzzy match.
    How matching works:
    – first characters of both `str1` and `str2` *must* match
    – `str1` length larger than `str2` length is allowed only when `unmatched` is true
    – ideal match is when `str1` equals to `str2` (score: 1)
    – next best match is `str2` starts with `str1` (score: 1 × percent of matched characters)
    – other scores depend on how close characters of `str1` to the beginning of `str2`
    :param partial_match Allow length `str1` to be greater than `str2` length
    """
    str1 = str1.lower()
    str2 = str2.lower()

    if str1 == str2:
        return 1

    # Both strings MUST start with the same character
    if not str1 or not str2 or str1[0] != str2[0]:
        return 0

    str1_len = len(str1)
    str2_len = len(str2)

    if not partial_match and str1_len > str2_len:
        return 0

    # Characters from `str1` which are closer to the beginning of a `str2` should
    # have higher score.
    # For example, if `str2` is `abcde`, it’s max score is:
    # 5 + 4 + 3 + 2 + 1 = 15 (sum of character positions in reverse order)
    # Matching `abd` against `abcde` should produce:
    # 5 + 4 + 2 = 11
    # Acronym bonus for match right after `-`. Matching `abd` against `abc-de`
    # should produce:
    # 6 + 5 + 4 (use `d` position in `abd`, not in abc-de`)

    min_length = min(str1_len, str2_len)
    max_length = max(str1_len, str2_len)
    i = 1
    j = 1
    score = max_length
    found = False
    acronym = False

    while i < str1_len:
        ch1 = str1[i]
        found = False
        acronym = False

        while j < str2_len:
            ch2 = str2[j]

            if ch1 == ch2:
                found = True
                score += max_length - (i if acronym else j)
                break

            # add acronym bonus for exactly next match after unmatched `-`
            acronym = ch2 == '-'
            j += 1


        if not found:
            if not partial_match:
                return 0
            break

        i += 1

    match_ratio = i / max_length
    delta = max_length - min_length
    max_score = n_sum(max_length) - n_sum(delta)
    return (score * match_ratio) / max_score


def n_sum(n: int):
    "Calculates sum of first `n` numbers, e.g. 1+2+3+...n"
    return n * (n + 1) / 2



from ..config import Config
from ..abbreviation import parse as abbreviation, Abbreviation, AbbreviationNode, AbbreviationAttribute
from .attributes import merge_attributes as attributes
from .snippets import resolve_snippets as snippets
from .implicit_tag import implicit_tag
from .lorem import lorem
from .addon.jsx import jsx
from .addon.xsl import xsl
from .addon.bem import bem
from .format import html, haml, slim, pug
from .utils import walk

FORMATTERS = {
    'html': html,
    'haml': haml,
    'slim': slim,
    'pug': pug
}

def parse(abbr: str, config: Config):
    """
    Parses given Emmet abbreviation into a final abbreviation tree with all
    required transformations applied
    """
    if isinstance(abbr, str):
        abbr = abbreviation(abbr, {
            'text': config.get('text'),
            'variables': config.variables,
            'max_repeat': config.get('maxRepeat') or config.get('max_repeat'),
            'jsx': bool(config.options.get('jsx.enabled'))
        })

    # Run abbreviation resolve in two passes:
    # 1. Map each node to snippets, which are abbreviations as well. A single snippet
    # may produce multiple nodes
    # 2. Transform every resolved node
    snippets(abbr, config)
    walk(abbr, transform, config)
    return abbr

def stringify(abbr: Abbreviation, config: Config):
    "Converts given abbreviation to string according to provided `config`"
    global FORMATTERS
    formatter = FORMATTERS.get(config.syntax, html)
    return formatter(abbr, config)


def transform(node: AbbreviationNode, ancestors: list, config: Config):
    "Modifies given node and prepares it for output"
    implicit_tag(node, ancestors, config)
    attributes(node, config)
    lorem(node, ancestors, config)

    if config.syntax == 'xsl':
        xsl(node)

    if config.options.get('jsx.enabled'):
        jsx(node)

    if config.options.get('bem.enabled'):
        bem(node, ancestors, config)



class Scanner:
    __slots__ = ('string', 'pos', 'start', 'end')

    def __init__(self, source: str, start=0, end=None):
        self.string = source
        self.pos = self.start = start
        self.end = len(source) if end is None else end

    def eof(self):
        "Returns true only if the stream is at the end of the file."
        return self.pos >= self.end

    def limit(self, start=0, end: int=None):
        """
        Creates a new stream instance which is limited to given `start` and `end`
        range. E.g. its `eof()` method will look at `end` property, not actual
        stream end
        """
        return Scanner(self.string, start, end)

    def peek(self):
        "Returns the next character in the stream without advancing it."
        return self.string[self.pos] if self.pos < self.end else ''

    def next(self):
        """
        Returns the next character in the stream and advances it.
        Returns `None` when no more characters left
        """
        if self.pos < self.end:
            ch = self.string[self.pos]
            self.pos += 1
            return ch

    def eat(self, match):
        """
        `match` can be a character or a function that takes a character
        and returns a boolean. If the next character in the stream 'matches'
        the given argument, it is consumed and returned.
        Otherwise, `False` is returned.
        """
        ch = self.peek()
        ok = match(ch) if callable(match) else match == ch
        if ok:
            self.pos += 1

        return ok

    def eat_while(self, match):
        """
        Repeatedly calls `eat` with the given argument, until it
        fails. Returns `True` if any characters were eaten.
        """
        start = self.pos
        while self.pos < self.end and self.eat(match):
            pass
        return self.pos != start

    def back_up(self, n: int = 1):
        """
        Backs up the stream n characters. Backing it up further than the
        start of the current token will cause things to break, so be careful.
        """
        self.pos -= n

    def current(self):
        "Get the string between the start of the current token and the current stream position."
        return self.substring(self.start, self.pos)

    def substring(self, start: int, end: int = None):
        "Returns substring for given range"
        if end is None:
            end = self.end
        return self.string[start:end]

    def error(self, message: str, pos: int = None):
        "Creates error object with current stream state"
        if pos is None:
            pos = self.pos
        return ScannerException("%s at %d" % (message, pos + 1), pos, self.string)


class ScannerException(Exception):
    def __init__(self, message: str, pos: int, source: str):
        super(ScannerException, self).__init__()
        self.message = message
        self.string = source
        self.pos = pos



from ..scanner_utils import is_space


class SelectItemModel:
    __slots__ = ('start', 'end', 'ranges')

    def __init__(self, start: int, end: int, ranges: list=None):
        self.start = start
        self.end = end
        self.ranges = ranges

    def to_json(self):
        return {
            'start': self.start,
            'end': self.end,
            'ranges': self.ranges
        }


def push_range(ranges: list, rng: list):
    prev = ranges and ranges[-1]
    if rng and rng[0] != rng[1] and (not prev or prev[0] != rng[0] or prev[1] != rng[1]):
        ranges.append(rng)


def token_list(value: str, offset=0):
    "Returns ranges of tokens in given value. Tokens are space-separated words."
    ranges = []
    l = len(value)
    pos = 0
    start = 0
    end = 0

    while pos < l:
        end = pos
        ch = value[pos]
        pos += 1
        if is_space(ch):
            if start != end:
                ranges.append((offset + start, offset + end))

            while pos < l and is_space(value[pos]):
                pos += 1

            start = pos

    if start != pos:
        ranges.append((offset + start, offset + pos))

    return ranges



from .utils import ScannerOptions, ElementType
from .scan import scan
from .attributes import attributes, AttributeToken

class MatchedTag:
    __slots__ = ('name', 'attributes', 'open', 'close')

    def __init__(self, name: str, attrs: list, open_range: tuple, close_range: tuple=None):
        self.name = name
        "Name of matched tag"

        self.attributes = attrs
        "List of tag attributes"

        self.open = open_range
        "Range of opening tag"

        self.close = close_range
        "Range of closing tag. If absent, tag is self-closing"


class Tag:
    __slots__ = ('name', 'start', 'end')

    def __init__(self, name: str, start: int, end: int):
        self.name = name
        self.start = start
        self.end = end


class InwardTag:
    __slots__ = ('name', 'ranges', 'first_child')

    def __init__(self, name: str, ranges: list, first_child=None):
        self.name = name
        self.ranges = ranges
        self.first_child = first_child


class BalancedTag:
    __slots__ = ('name', 'open', 'close')

    def __init__(self, name: str, open_range: tuple, close_range: tuple=None):
        self.name = name
        "Name of matched tag"

        self.open = open_range
        "Range of opening tag"

        self.close = close_range
        "Range of closing tag. If absent, tag is self-closing"

    def to_json(self):
        json = {
            'name': self.name,
            'open': list(self.open)
        }
        if self.close:
            json['close'] = list(self.close)

        return json


def match(source: str, pos: int, opt: dict=None) -> MatchedTag:
    "Finds matched tag for given `pos` location in XML/HTML `source`"
    # Since we expect large input document, we’ll use pooling technique
    # for storing tag data to reduce memory pressure and improve performance
    pool = []
    stack = []
    options = ScannerOptions(opt)
    result = [None]

    def scan_callback(name: str, elem_type: ElementType, start: int, end: int):
        if elem_type == ElementType.Open and is_self_close(name, options):
            # Found empty element in HTML mode, mark is as self-closing
            elem_type = ElementType.SelfClose

        if elem_type == ElementType.Open:
            # Allocate tag object from pool
            stack.append(alloc_tag(pool, name, start, end))
        elif elem_type == ElementType.SelfClose:
            if start < pos < end:
                # Matched given self-closing tag
                result[0] = MatchedTag(name, get_attributes(source, start, end, name), (start, end))
                return False
        else:
            tag = stack and stack[-1]
            if tag and tag.name == name:
                # Matching closing tag found
                if tag.start < pos < end:
                    result[0] = MatchedTag(name, get_attributes(source, tag.start, tag.end, name), (tag.start, tag.end), (start, end))
                    return False

                if stack:
                    # Release tag object for further re-use
                    release_tag(pool, stack.pop())

    scan(source, scan_callback, options.special)
    return result[0]


def balanced_outward(source: str, pos: int, opt: dict=None) -> list:
    """
    Returns balanced tag model: a list of all XML/HTML tags that could possibly match
    given location when moving in outward direction
    """
    pool = []
    stack = []
    options = ScannerOptions(opt)
    result = []

    def scan_callback(name: str, elem_type: ElementType, start: int, end: int):
        if elem_type == ElementType.Close:
            tag = stack and stack[-1]
            if tag and tag.name == name:
                # XXX check for invalid tag names?
                # Matching closing tag found, check if matched pair is a candidate
                # for outward balancing
                if tag.start < pos < end:
                    result.append(BalancedTag(name, (tag.start, tag.end), (start, end)))
                # Release tag object for further re-use
                release_tag(pool, stack.pop())
        elif elem_type == ElementType.SelfClose or is_self_close(name, options):
            if start < pos < end:
                # Matched self-closed tag
                result.append(BalancedTag(name, (start, end)))
        else:
            stack.append(alloc_tag(pool, name, start, end))

    scan(source, scan_callback, options.special)
    return result


def balanced_inward(source: str, pos: int, opt: dict=None) -> list:
    """
    Returns balanced tag model: a list of all XML/HTML tags that could possibly match
    given location when moving in inward direction
    """
    # Collecting tags for inward balancing is a bit trickier: we have to store
    # first child of every matched tag until we find the one that matches given
    # location
    pool = []
    stack = []
    options = ScannerOptions(opt)
    result = []

    def alloc(name: str, start: int, end: int):
        if pool:
            tag = pool.pop()
            tag.name = name
            tag.ranges.append(start)
            tag.ranges.append(end)
            return tag

        return InwardTag(name, [start, end])

    def release(tag: InwardTag):
        tag.ranges.clear()
        tag.first_child = None
        pool.append(tag)

    def scan_callback(name: str, elem_type: ElementType, start: int, end: int):
        if elem_type == ElementType.Close:
            if not stack:
                # Some sort of lone closing tag, ignore it
                return

            tag = stack[-1]
            if tag.name == name:
                # XXX check for invalid tag names?
                # Matching closing tag found, check if matched pair is a candidate
                # for outward balancing
                if tag.ranges[0] <= pos <= end:
                    result.append(BalancedTag(name, (tag.ranges[0], tag.ranges[1]), (start, end)))

                    while tag.first_child:
                        child = tag.first_child
                        res = BalancedTag(child.name, (child.ranges[0], child.ranges[1]))
                        if len(child.ranges) > 2:
                            res.close = (child.ranges[2], child.ranges[3])
                        result.append(res)
                        release(tag)
                        tag = child

                    return False
                else:
                    stack.pop()
                    parent = stack and stack[-1]
                    if parent and not parent.first_child:
                        # No first child in parent node: store current tag
                        tag.ranges.append(start)
                        tag.ranges.append(end)
                        parent.first_child = tag
                    else:
                        release(tag)
        elif elem_type == ElementType.SelfClose or is_self_close(name, options):
            if start < pos < end:
                # Matched self-closed tag, no need to look further
                result.append(BalancedTag(name, (start, end)))
                return False

            parent = stack and stack[-1]
            if parent and not parent.first_child:
                parent.first_child = alloc(name, start, end)
        else:
            stack.append(alloc(name, start, end))

    scan(source, scan_callback, options.special)
    return result


def alloc_tag(pool: list, name: str, start: int, end: int):
    if pool:
        tag = pool.pop()
        tag.name = name
        tag.start = start
        tag.end = end
        return tag
    return Tag(name, start, end)

def release_tag(pool: list, tag: Tag):
    if tag:
        pool.append(tag)


def get_attributes(source: str, start: int, end: int, name: str=None):
    "Returns parsed attributes from given source"
    attrs = attributes(source[start:end], name)
    for attr in attrs:
        attr.name_start += start
        attr.name_end += start
        if attr.value is not None:
            attr.value_start += start
            attr.value_end += start

    return attrs


def is_self_close(name: str, options: ScannerOptions):
    "Check if given tag is self-close for current parsing context"
    return not options.xml and name in options.empty



from ...scanner import Scanner
from ...scanner_utils import is_quote, is_space, is_number, is_alpha, is_alpha_numeric_word
from .utils import Chars, escaped
from . import tokens

OPERATOR_TYPES = dict([
    (Chars.Child, 'child'),
    (Chars.Sibling, 'sibling'),
    (Chars.Climb, 'climb'),
    (Chars.Dot, 'class'),
    (Chars.Hash, 'id'),
    (Chars.Slash, 'close'),
    (Chars.Equals, 'equal')
])


def tokenize(source: str):
    scanner = Scanner(source)
    result = []
    ctx = {
        'group': 0,
        'attribute': 0,
        'expression': 0,
        'quote': None
    }

    while not scanner.eof():
        ch = scanner.peek()
        token = field(scanner, ctx) or \
            repeater_placeholder(scanner) or \
            repeater_number(scanner) or \
            repeater(scanner) or \
            white_space(scanner) or \
            literal(scanner, ctx) or \
            operator(scanner) or \
            quote(scanner) or \
            bracket(scanner)

        if token:
            result.append(token)
            if token.type == 'Quote':
                ctx['quote'] = None if ch == ctx['quote'] else ch
            elif token.type == 'Bracket':
                ctx[token.context] += 1 if token.open else -1
        else:
            raise scanner.error('Unexpected character')

    return result


def literal(scanner: Scanner, ctx: dict):
    "Consumes literal from given scanner"
    start = scanner.pos
    value = []

    while not scanner.eof():
        if escaped(scanner):
            value.append(scanner.current())
            continue

        ch = scanner.peek()

        if ch == ctx['quote'] or ch == Chars.Dollar or is_allowed_operator(ch, ctx):
            # 1. Found matching quote
            # 2. The `$` character has special meaning in every context
            # 3. Depending on context, some characters should be treated as operators
            break

        if ctx['expression'] and ch == Chars.CurlyBracketClose:
            break

        if not ctx['quote'] and not ctx['expression']:
            # Consuming element name
            if not ctx['attribute'] and not is_element_name(ch):
                break

            if is_allowed_space(ch, ctx) or \
               is_allowed_repeater(ch, ctx) or \
               is_quote(ch) or \
               bracket_type(ch):
                # Stop for characters not allowed in unquoted literal
                break

        value.append(scanner.next())

    if start != scanner.pos:
        scanner.start = start
        return tokens.Literal(''.join(value), start, scanner.pos)


def white_space(scanner: Scanner):
    "Consumes white space characters as string literal from given scanner"
    start = scanner.pos
    if scanner.eat_while(is_space):
        return tokens.WhiteSpace(start, scanner.pos)


def quote(scanner: Scanner):
    "Consumes quote from given scanner"
    ch = scanner.peek()
    if is_quote(ch):
        return tokens.Quote(ch == Chars.SingleQuote, inc_pos(scanner), scanner.pos)


def bracket(scanner: Scanner):
    "Consumes bracket from given scanner"
    ch = scanner.peek()
    context = bracket_type(ch)
    if context:
        return tokens.Bracket(is_open_bracket(ch), context, inc_pos(scanner), scanner.pos)


def operator(scanner: Scanner):
    "Consumes operator from given scanner"
    op = operator_type(scanner.peek())
    if op:
        return tokens.Operator(op, inc_pos(scanner), scanner.pos)


def repeater(scanner: Scanner):
    "Consumes node repeat token from current scanner position and returns its parsed value"
    start = scanner.pos
    if scanner.eat(Chars.Asterisk):
        scanner.start = scanner.pos
        count = 1
        implicit = False

        if scanner.eat_while(is_number):
            count = int(scanner.current())
        else:
            implicit = True

        return tokens.Repeater(count, 0, implicit, start, scanner.pos)


def repeater_placeholder(scanner: Scanner):
    "Consumes repeater placeholder `$#` from given scanner"
    start = scanner.pos
    if scanner.eat(Chars.Dollar) and scanner.eat(Chars.Hash):
        return tokens.RepeaterPlaceholder(None, start, scanner.pos)

    scanner.pos = start


def repeater_number(scanner: Scanner):
    "Consumes numbering token like `$` from given scanner state"
    start = scanner.pos
    if scanner.eat_while(Chars.Dollar):
        size = scanner.pos - start
        reverse = False
        base = 1
        parent = 0

        if scanner.eat(Chars.At):
            # Consume numbering modifiers
            while scanner.eat(Chars.Climb):
                parent += 1

            reverse = scanner.eat(Chars.Dash)
            scanner.start = scanner.pos
            if scanner.eat_while(is_number):
                base = int(scanner.current())

        scanner.start = start
        return tokens.RepeaterNumber(size, reverse, base, parent, start, scanner.pos)


def field(scanner: Scanner, ctx: dict):
    start = scanner.pos
    # Fields are allowed inside expressions and attributes
    if (ctx['expression'] or ctx['attribute']) and scanner.eat(Chars.Dollar) and scanner.eat(Chars.CurlyBracketOpen):
        scanner.start = scanner.pos
        index = None
        name = ''

        if scanner.eat_while(is_number):
            # It’s a field
            index = int(scanner.current())
            if scanner.eat(Chars.Colon):
                name = consume_placeholder(scanner)
        elif is_alpha(scanner.peek()):
            # It’s a variable
            name = consume_placeholder(scanner)

        if scanner.eat(Chars.CurlyBracketClose):
            return tokens.Field(name, index, start, scanner.pos)

        raise scanner.error('Expecting }')


    # If we reached here then there’s no valid field here, revert
    # back to starting position
    scanner.pos = start


def consume_placeholder(scanner: Scanner):
    "Consumes a placeholder: value right after `:` in field. Could be empty"
    stack = []
    scanner.start = scanner.pos

    while not scanner.eof():
        if scanner.eat(Chars.CurlyBracketOpen):
            stack.append(scanner.pos)
        elif scanner.eat(Chars.CurlyBracketClose):
            if not len(stack):
                scanner.pos -= 1
                break
            stack.pop()
        else:
            scanner.pos += 1

    if len(stack):
        scanner.pos = stack.pop()
        raise scanner.error('Expecting }')

    return scanner.current()


def is_allowed_operator(ch: str, ctx: dict):
    "Check if given character code is an operator and it’s allowed in current context"
    op = operator_type(ch)
    if not op or ctx['quote'] or ctx['expression']:
        # No operators inside quoted values or expressions
        return False

    # Inside attributes, only `equals` is allowed
    return not ctx['attribute'] or op == 'equal'


def is_allowed_space(ch: str, ctx: dict):
    """
    Check if given character is a space character and is allowed to be consumed
    as a space token in current context
    """
    return is_space(ch) and not ctx['expression']


def is_allowed_repeater(ch: str, ctx: dict):
    "Check if given character can be consumed as repeater in current context"
    return ch == Chars.Asterisk and not ctx['attribute'] and not ctx['expression']


def bracket_type(ch: str):
    "If given character is a bracket, returns it’s type"
    if ch in (Chars.RoundBracketOpen, Chars.RoundBracketClose):
        return 'group'

    if ch in (Chars.SquareBracketOpen, Chars.SquareBracketClose):
        return 'attribute'

    if ch in (Chars.CurlyBracketOpen, Chars.CurlyBracketClose):
        return 'expression'

def operator_type(ch: int):
    "If given character is an operator, returns it’s type"
    global OPERATOR_TYPES
    return OPERATOR_TYPES.get(ch)


def is_open_bracket(ch: str):
    "Check if given character is an open bracket"
    return ch in (Chars.CurlyBracketOpen, Chars.SquareBracketOpen, Chars.RoundBracketOpen)

def is_element_name(ch: str):
    "Check if given character is allowed in element name"
    return is_alpha_numeric_word(ch) or ch in (Chars.Dash, Chars.Colon, Chars.Excl)

def inc_pos(scanner: Scanner):
    pos = scanner.pos
    scanner.pos += 1
    return pos



from ..css_matcher import scan, split_value, TokenType
from .utils import push_range, SelectItemModel

class CSSSection:
    __slots__ = ('start', 'end', 'body_start', 'body_end', 'properties')

    def __init__(self, start: int, end: int, body_start: int, body_end: int, properties: list=None):
        self.start = start
        self.end = end
        self.body_start = body_start
        self.body_end = body_end
        self.properties = properties

    def to_json(self):
        result = {
            'start': self.start,
            'end': self.end,
            'body_start': self.body_start,
            'body_end': self.body_end
        }

        if self.properties:
            result['properties'] = [prop.to_json() for prop in self.properties]

        return result


class CSSProperty:
    __slots__ = ('name', 'value', 'value_tokens', 'before', 'after')

    def __init__(self, code: str, name: list, before: int, start: int, end: int, delimiter: int, offset=0):
        self.name = (offset + name[0], offset + name[1])
        self.value = (offset + start, offset + end)
        self.value_tokens = split_value(code[start:end], offset + start)
        self.before = before
        self.after = offset + delimiter + 1

    def to_json(self):
        return {
            'name': self.name,
            'value': self.value,
            'value_tokens': self.value_tokens,
            'before': self.before,
            'after': self.after
        }


class ParseState:
    __slots__ = ('type', 'start', 'end', 'value_start', 'value_end', 'value_delimiter')

    def __init__(self):
        self.type = None
        self.start = -1
        self.end = -1
        self.value_start = -1
        self.value_end = -1
        self.value_delimiter = -1


def get_css_section(code: str, pos: int, properties=False) -> CSSSection:
    """
    Returns context CSS section for given location in source code
    :param properties Parse inner properties
    """
    stack = []
    pool = []
    result = []
    result.append(None) # Skip pylint warnings

    def scan_callback(token_type: str, start: int, end: int, delimiter: int):
        if start > pos and not stack:
            return False

        if token_type == TokenType.Selector:
            stack.append(alloc_range(pool, start, end, delimiter))
        elif token_type == TokenType.BlockEnd:
            sel = stack and stack.pop()
            if sel and sel[0] <= pos <= end:
                result[0] = CSSSection(sel[0], end, sel[2] + 1, start)
                return False
            release_range(pool, sel)


    scan(code, scan_callback)
    section = result[0]

    if section and properties:
        section.properties = parse_properties(code, section.body_start, section.body_end)

    return section


def select_item_css(code: str, pos: int, is_prev=False) -> SelectItemModel:
    "Returns list of ranges for Select Next/Previous CSS Item  action"
    if is_prev:
        return select_previous_item(code, pos)
    return select_next_item(code, pos)


def select_next_item(code: str, pos: int) -> SelectItemModel:
    "Returns regions for selecting next item in CSS"
    result = []
    result.append(None)
    pending_property = []
    pending_property.append(None)

    def scan_callback(token_type: str, start: int, end: int, delimiter: int):
        if start < pos:
            return

        if token_type == TokenType.Selector:
            result[0] = SelectItemModel(start, end, [(start, end)])
            return False
        elif token_type == TokenType.PropertyName:
            pending_property[0] = (start, end, delimiter)
        elif token_type == TokenType.PropertyValue:
            section = SelectItemModel(start, delimiter + 1 if delimiter != -1 else end, [])
            result[0] = section

            if pending_property[0]:
                # Full property range
                prop = pending_property[0]
                section.start = prop[0]
                push_range(section.ranges, (prop[0], section.end))


            # Full value range
            push_range(section.ranges, (start, end))

            # Value fragments
            for r in split_value(code[start:end]):
                push_range(section.ranges, (r[0] + start, r[1] + start))

            return False
        elif pending_property[0]:
            prop = pending_property[0]
            result[0] = SelectItemModel(prop[0], prop[1], [(prop[0], prop[1])])
            return False

    scan(code, scan_callback)
    return result[0]


def select_previous_item(code: str, pos: int) -> SelectItemModel:
    "Returns regions for selecting previous item in CSS"
    state = ParseState()

    def scan_callback(token_type, start, end, delimiter):
        # Accumulate context until we reach given position
        if start >= pos and token_type != TokenType.PropertyValue:
            return False

        if token_type in (TokenType.Selector, TokenType.PropertyName):
            state.start = start
            state.end = end
            state.type = token_type
            state.value_start = state.value_end = state.value_delimiter = -1
        elif token_type == TokenType.PropertyValue:
            state.value_start = start
            state.value_end = end
            state.value_delimiter = delimiter

    scan(code, scan_callback)

    if state.type == TokenType.Selector:
        return SelectItemModel(state.start, state.end, [(state.start, state.end)])

    if state.type == TokenType.PropertyName:
        result = SelectItemModel(state.start, state.end, [])

        if state.value_start != -1:
            result.end = state.value_delimiter + 1 if state.value_delimiter != -1 else state.value_end
            # Full property range
            push_range(result.ranges, (state.start, result.end))

            # Full value range
            push_range(result.ranges, (state.value_start, state.value_end))

            # Value fragments
            for r in split_value(code[state.value_start:state.value_end]):
                push_range(result.ranges, (r[0] + state.value_start, r[1] + state.value_start))
        else:
            push_range(result.ranges, (state.start, state.end))

        return result


class ParsePropertiesState:
    __slots__ = ('pending_name', 'nested', 'before')

    def __init__(self, before: int):
        self.pending_name = None
        self.nested = 0
        self.before= before

def parse_properties(code: str, parse_from=0, parse_to=None) -> list:
    """
    Parses properties in `from:to` fragment of `code`. Note that `from:to` must
    point to CSS section content, e.g. *inside* `{` and `}` (or top-level code context),
    all properties found in nested sections will be ignored
    """
    if parse_to is None:
        parse_to = len(code)

    fragment = code[parse_from:parse_to]
    result = []
    pool = []
    state = ParsePropertiesState(parse_from)

    def scan_callback(token_type, start: int, end: int, delimiter: int):
        if token_type == TokenType.Selector:
            state.nested += 1
        elif token_type == TokenType.BlockEnd:
            state.nested -= 1
            state.before = parse_from + end
        elif not state.nested:
            if token_type == TokenType.PropertyName:
                if state.pending_name:
                    # Create property with empty value
                    value_pos = state.pending_name[2]
                    result.append(
                        CSSProperty(fragment, state.pending_name, state.before,
                                    value_pos, value_pos, value_pos,
                                    parse_from))
                    release_range(pool, state.pending_name)
                    state.before = parse_from + start
                state.pending_name = alloc_range(pool, start, end, delimiter)
            elif token_type == TokenType.PropertyValue:
                if state.pending_name:
                    result.append(
                        CSSProperty(fragment, state.pending_name, state.before,
                                    start, end, delimiter, parse_from))
                    release_range(pool, state.pending_name)
                    state.pending_name = None
                state.before = parse_from + delimiter + 1

    scan(fragment, scan_callback)
    return result


def alloc_range(pool: list, start: int, end: int, delimiter: int) -> list:
    "Allocates new token range from pool"
    if pool:
        rng = pool.pop()
        rng[0] = start
        rng[1] = end
        rng[2] = delimiter
        return rng

    return [start, end, delimiter]


def release_range(pool: list, rng: list):
    "Releases given token range and pushes it back into the pool"
    if rng:
        pool.append(rng)



class BackwardScanner:
    __slots__ = ('text', 'start', 'pos')

    def __init__(self, text: str, start=0):
        self.text = text
        "Text to scan"

        self.start = start
        "Left bound till given text must be scanned"

        self.pos = len(text)
        "Current scanner position"

    def sol(self):
        "Check if given scanner position is at start of scanned text"
        return self.pos == self.start

    def peek(self, offset=0):
        "“Peeks” character code an current scanner location without advancing it"
        pos = self.pos - 1 + offset
        return self.text[pos] if 0 <= pos < len(self.text) else ''

    def previous(self):
        "Returns current character code and moves character location one symbol back"
        if not self.sol():
            self.pos -= 1
            return self.text[self.pos]

    def consume(self, match: callable):
        "Consumes current character code if it matches given `match` code or function"
        if self.sol():
            return False

        ok = match(self.peek()) if callable(match) else match == self.peek()
        if ok: self.pos -= 1

        return bool(ok)

    def consume_while(self, match: callable):
        start = self.pos
        while self.consume(match): pass
        return self.pos < start



vocabulary = {
	"common": ["далеко-далеко", "за", "словесными", "горами", "в стране", "гласных", "и согласных", "живут", "рыбные", "тексты"],
	"words": ["вдали", "от всех", "они", "буквенных", "домах", "на берегу", "семантика",
		"большого", "языкового", "океана", "маленький", "ручеек", "даль",
		"журчит", "по всей", "обеспечивает", "ее","всеми", "необходимыми",
		"правилами", "эта", "парадигматическая", "страна", "которой", "жаренные",
		"предложения", "залетают", "прямо", "рот", "даже", "всемогущая",
		"пунктуация", "не", "имеет", "власти", "над", "рыбными", "текстами",
		"ведущими", "безорфографичный", "образ", "жизни", "однажды", "одна",
		"маленькая", "строчка","рыбного", "текста", "имени", "lorem", "ipsum",
		"решила", "выйти", "большой", "мир", "грамматики", "великий", "оксмокс",
		"предупреждал", "о", "злых", "запятых", "диких", "знаках", "вопроса",
		"коварных", "точках", "запятой", "но", "текст", "дал", "сбить",
		"себя", "толку", "он", "собрал", "семь", "своих", "заглавных", "букв",
		"подпоясал", "инициал", "за", "пояс", "пустился", "дорогу",
		"взобравшись", "первую", "вершину", "курсивных", "гор", "бросил",
		"последний", "взгляд", "назад", "силуэт", "своего", "родного", "города",
		"буквоград", "заголовок", "деревни", "алфавит", "подзаголовок", "своего",
		"переулка", "грустный", "реторический", "вопрос", "скатился", "его",
		"щеке", "продолжил", "свой", "путь", "дороге", "встретил", "рукопись",
		"она", "предупредила",  "моей", "все", "переписывается", "несколько",
		"раз", "единственное", "что", "меня", "осталось", "это", "приставка",
		"возвращайся", "ты", "лучше", "свою", "безопасную", "страну", "послушавшись",
		"рукописи", "наш", "продолжил", "свой", "путь", "вскоре", "ему",
		"повстречался", "коварный", "составитель", "рекламных", "текстов",
		"напоивший", "языком", "речью", "заманивший", "свое", "агентство",
		"которое", "использовало", "снова", "снова", "своих", "проектах",
		"если", "переписали", "то", "живет", "там", "до", "сих", "пор"]
}



from ..scanner import Scanner
from ..scanner_utils import is_quote, is_space

class ScanState:
    __slots__ = ('start', 'end', 'property_delimiter', 'property_start',
        'property_end', 'expression')

    def __init__(self):
        self.start = -1
        "Start location of currently consumed token"

        self.end = -1
        "End location of currently consumed token"

        self.property_delimiter = -1
        "Location of possible property delimiter"

        self.property_start = -1
        "Location of possible property start"

        self.property_end = -1
        "Location of possible property end"

        self.expression = 0
        "In expression context"

    def reset(self):
        self.start = self.end = self.property_start = self.property_end = self.property_delimiter = -1



class TokenType:
    Selector = 'selector'
    PropertyName = 'propertyName'
    PropertyValue = 'propertyValue'
    BlockEnd = 'blockEnd'


class Chars:
    LeftCurly = '{'
    RightCurly = '}'
    Asterisk = '*'
    Slash = '/'
    Colon = ':'
    Semicolon = ';'
    Backslash = '\\'
    LeftRound = '('
    RightRound = ')'
    LF = '\n'
    CR = '\r'

def scan(source: str, callback: callable):
    """
    Performs fast scan of given stylesheet (CSS, LESS, SCSS) source code and runs
    callback for each token and its range found. The goal of this parser is to quickly
    determine document structure: selector, property, value and block end.
    It doesn’t provide detailed info about CSS atoms like compound selectors,
    operators, quoted string etc. to reduce memory allocations: this data can be
    parsed later on demand.
    """
    scanner = Scanner(source)
    state = ScanState()
    block_end = False

    def notify(token_type: TokenType, delimiter: int=None, start: int=None, end: int=None):
        if delimiter is None: delimiter = scanner.start
        if start is None: start = state.start
        if end is None: end = state.end
        return callback(token_type, start, end, delimiter) is False

    while not scanner.eof():
        if comment(scanner) or whitespace(scanner):
            continue

        scanner.start = scanner.pos
        block_end = scanner.eat(Chars.RightCurly)
        if block_end or scanner.eat(Chars.Semicolon):
            # Block or property end
            if state.property_start != -1:
                # We have pending property
                if notify(TokenType.PropertyName, state.property_delimiter, state.property_start, state.property_end):
                    return

                if state.start == -1:
                    # Explicit property value state: emit empty value
                    state.start = state.end = scanner.start

                if notify(TokenType.PropertyValue):
                    return
            elif state.start != -1 and notify(TokenType.PropertyName):
                # Flush consumed token
                return

            if block_end:
                state.start = scanner.start
                state.end = scanner.pos

                if notify(TokenType.BlockEnd):
                    return

            state.reset()
        elif scanner.eat(Chars.LeftCurly):
            # Block start
            if state.start == -1 and state.property_start == -1:
                # No consumed selector, emit empty value as selector start
                state.start = state.end = scanner.pos

            if state.property_start != -1:
                # Now we know that value that looks like property name-value pair
                # was actually a selector
                state.start = state.property_start

            if notify(TokenType.Selector):
                return
            state.reset()
        elif scanner.eat(Chars.Colon) and not is_known_selector_colon(scanner, state):
            # Colon could be one of the following:
            # — property delimiter: `foo: bar`, must be in block context
            # — variable delimiter: `$foo: bar`, could be anywhere
            # — pseudo-selector: `a:hover`, could be anywhere (for LESS and SCSS)
            # — media query expression: `min-width: 100px`, must be inside expression context
            # Since I can’t easily detect `:` meaning for sure, we’ll update state
            # to accumulate possible property name-value pair or selector
            if state.property_start == -1:
                state.property_start = state.start
            state.property_end = state.end
            state.property_delimiter = scanner.pos - 1
            state.start = state.end = -1
        else:
            if state.start == -1:
                state.start = scanner.pos

            if scanner.eat(Chars.LeftRound):
                state.expression += 1
            elif scanner.eat(Chars.RightRound):
                state.expression -= 1
            elif not literal(scanner):
                scanner.pos += 1

            state.end = scanner.pos

    if state.property_start != -1:
        # Pending property name
        if notify(TokenType.PropertyName, state.property_delimiter, state.property_start, state.property_end):
            return

    if state.start != -1:
        # There’s pending token in state
        notify(TokenType.PropertyValue if state.property_start != -1 else TokenType.PropertyName, -1)


def whitespace(scanner: Scanner):
    return scanner.eat_while(is_space)


def comment(scanner: Scanner):
    """
    Consumes CSS comments from scanner: `/*  * /`
    It’s possible that comment may not have closing part
    """
    start = scanner.pos
    if scanner.eat(Chars.Slash) and scanner.eat(Chars.Asterisk):
        scanner.start = start
        while not scanner.eof():
            if scanner.eat(Chars.Asterisk):
                if scanner.eat(Chars.Slash):
                    return True
                continue
            scanner.pos += 1
        return True
    else:
        scanner.pos = start

    return False


def literal(scanner: Scanner):
    ch = scanner.peek()
    if is_quote(ch):
        scanner.start = scanner.pos
        scanner.pos += 1
        while not scanner.eof():
            if scanner.eat(ch) or scanner.eat(Chars.LF) or scanner.eat(Chars.CR):
                break

            # Skip escape character, if any
            scanner.eat(Chars.Backslash)
            scanner.pos += 1

        # Do not throw if string is incomplete
        return True


def is_known_selector_colon(scanner: Scanner, state: ScanState):
    "Check if current state is a known selector context for `:` delimiter"
    # Either inside expression like `(min-width: 10px)` or pseudo-element `::before`
    return state.expression or scanner.eat_while(Chars.Colon)



from math import floor
from .parser import parse, TokenType, Operator, Token, MathExpressionException
from .extract import extract

ops1 = {
    Operator.Minus: lambda num: -num
}

ops2 = {
    Operator.Plus: lambda a, b: a + b,
    Operator.Minus: lambda a, b: a - b,
    Operator.Multiply: lambda a, b: a * b,
    Operator.Divide: lambda a, b: a / b,
    Operator.IntDivide: lambda a, b: floor(a / b)
}

def evaluate(expr: str):
    if not isinstance(expr, (list, tuple)):
        expr = parse(expr)

    if not expr:
        return None

    n_stack = []

    for token in expr:
        if token.type == TokenType.Number:
            n_stack.append(token.value)
        elif token.type == TokenType.Op2:
            n2 = n_stack.pop()
            n1 = n_stack.pop()
            f = ops2[token.value]
            n_stack.append(f(n1, n2))
        elif token.type == TokenType.Op1:
            n1 = n_stack.pop()
            f = ops1[token.value]
            n_stack.append(f(n1))
        else:
            raise Exception('Invalid expression')

    if len(n_stack) > 1:
        raise Exception('Invalid Expression (parity)')

    return n_stack[0]



vocabulary = {
	"common": ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipisicing", "elit"],
	"words": ["exercitationem", "perferendis", "perspiciatis", "laborum", "eveniet",
		"sunt", "iure", "nam", "nobis", "eum", "cum", "officiis", "excepturi",
		"odio", "consectetur", "quasi", "aut", "quisquam", "vel", "eligendi",
		"itaque", "non", "odit", "tempore", "quaerat", "dignissimos",
		"facilis", "neque", "nihil", "expedita", "vitae", "vero", "ipsum",
		"nisi", "animi", "cumque", "pariatur", "velit", "modi", "natus",
		"iusto", "eaque", "sequi", "illo", "sed", "ex", "et", "voluptatibus",
		"tempora", "veritatis", "ratione", "assumenda", "incidunt", "nostrum",
		"placeat", "aliquid", "fuga", "provident", "praesentium", "rem",
		"necessitatibus", "suscipit", "adipisci", "quidem", "possimus",
		"voluptas", "debitis", "sint", "accusantium", "unde", "sapiente",
		"voluptate", "qui", "aspernatur", "laudantium", "soluta", "amet",
		"quo", "aliquam", "saepe", "culpa", "libero", "ipsa", "dicta",
		"reiciendis", "nesciunt", "doloribus", "autem", "impedit", "minima",
		"maiores", "repudiandae", "ipsam", "obcaecati", "ullam", "enim",
		"totam", "delectus", "ducimus", "quis", "voluptates", "dolores",
		"molestiae", "harum", "dolorem", "quia", "voluptatem", "molestias",
		"magni", "distinctio", "omnis", "illum", "dolorum", "voluptatum", "ea",
		"quas", "quam", "corporis", "quae", "blanditiis", "atque", "deserunt",
		"laboriosam", "earum", "consequuntur", "hic", "cupiditate",
		"quibusdam", "accusamus", "ut", "rerum", "error", "minus", "eius",
		"ab", "ad", "nemo", "fugit", "officia", "at", "in", "id", "quos",
		"reprehenderit", "numquam", "iste", "fugiat", "sit", "inventore",
		"beatae", "repellendus", "magnam", "recusandae", "quod", "explicabo",
		"doloremque", "aperiam", "consequatur", "asperiores", "commodi",
		"optio", "dolor", "labore", "temporibus", "repellat", "veniam",
		"architecto", "est", "esse", "mollitia", "nulla", "a", "similique",
		"eos", "alias", "dolore", "tenetur", "deleniti", "porro", "facere",
		"maxime", "corrupti"]
}



import re
from ...abbreviation import AbbreviationNode
from ...config import Config
from ...list_utils import get_item

class BEMData:
    __slots__ = ('class_names', 'block')

    def __init__(self, class_names: list, block: str=None):
        self.class_names = class_names
        self.block = block

re_element = re.compile(r'^(-+)([a-z0-9]+[a-z0-9-]*)', re.I)
re_modifier = re.compile(r'^(_+)([a-z0-9]+[a-z0-9-_]*)', re.I)

def block_candidates1(class_name: str):
    return re.match(r'^[a-z]-', class_name, re.I)

def block_candidates2(class_name: str):
    return re.match(r'^[a-z]', class_name, re.I)

def bem(node: AbbreviationNode, ancestors: list, config: Config):
    lookup = {}
    expand_class_names(node, lookup)
    expand_short_notation(node, ancestors, config, lookup)

def expand_class_names(node: AbbreviationNode, lookup: dict):
    """
    Expands existing class names in BEM notation in given `node`.
    For example, if node contains `b__el_mod` class name, this method ensures
    that element contains `b__el` class as well
    """
    data = get_bem_data(node, lookup)
    class_names = []

    for cl in data.class_names:
        # remove all modifiers and element prefixes from class name to get a base element name
        ix = cl.find('_')
        if ix > 0 and cl[0] != '-':
            class_names.append(cl[0:ix])
            class_names.append(cl[ix:])
        else:
            class_names.append(cl)

    if class_names:
        data.class_names = unique(class_names)
        data.block = find_block_name(data.class_names)
        update_class(node, ' '.join(data.class_names))


def expand_short_notation(node: AbbreviationNode, ancestors: list, config: Config, lookup: dict):
    data = get_bem_data(node, lookup)
    class_names = []
    options = config.options
    path = ancestors[1:] + [node]

    for cl in data.class_names:
        prefix = ''
        original_class = cl

        # parse element definition (could be only one)
        m = re_element.match(cl)
        if m:
            prefix = ''.join((get_block_name(path, len(m.group(1)), config.context) + options.get('bem.element') + m.group(2)))
            class_names.append(prefix)
            cl = cl[len(m.group(0)):]

        # parse modifiers definitions
        m = re_modifier.match(cl)
        if m:
            if not prefix:
                prefix = get_block_name(path, len(m.group(1)))
                class_names.append(prefix)

            class_names.append(''.join( (prefix, options.get('bem.modifier'), m.group(2)) ))
            cl = cl[len(m.group(0)):]

        if cl == original_class:
            # class name wasn’t modified: it’s not a BEM-specific class,
            # add it as-is into output
            class_names.append(original_class)

    arr_class_names = unique(class_names)
    if arr_class_names:
        update_class(node, ' '.join(arr_class_names))


def get_bem_data(node: AbbreviationNode, lookup: dict):
    "Returns BEM data from given abbreviation node"
    if node not in lookup:
        class_value = ''
        if node.attributes:
            for attr in node.attributes:
                if attr.name == 'class' and attr.value:
                    class_value = stringify_value(attr.value)
                    break

        lookup[node] = parse_bem(class_value)

    return lookup[node]


def get_bem_data_from_context(context: dict):
    # XXX dict is unhashable, can’t use lookup. Just re-parse on each request
    attrs = context.get('attributes', {})
    return parse_bem(attrs.get('class', ''))


def parse_bem(class_value=''):
    "Parses BEM data from given class name"
    class_names = class_value.split() if class_value else []
    return BEMData(class_names, find_block_name(class_names))


def get_block_name(ancestors: list, depth=0, context: dict=None, lookup={}):
    """
    Returns block name for given `node` by `prefix`, which tells the depth of
    of parent node lookup
    """
    max_parent_ix = 0
    parent_ix = max(len(ancestors) - depth, max_parent_ix)
    while max_parent_ix <= parent_ix:
        parent = get_item(ancestors, parent_ix)
        if parent:
            data = get_bem_data(parent, lookup)
            if data.block:
                return data.block
        parent_ix -= 1

    if context is not None:
        data = get_bem_data_from_context(context)
        if data.block:
            return data.block

    return ''


def find_block_name(class_names: list):
    return find(class_names, block_candidates1) or find(class_names, block_candidates2) or None


def find(class_names: list, fn: callable):
    for cl in class_names:
        if re_element.match(cl) or re_modifier.match(cl):
            break

        if fn(cl): return cl


def update_class(node: AbbreviationNode, value: str):
    for attr in node.attributes:
        if attr.name == 'class':
            attr.value = [value]
            break


def stringify_value(value: list):
    return ''.join([t if isinstance(t, str) else t.name for t in value])

def unique(items: list):
    seen = set()
    return [x for x in items if x not in seen and not seen.add(x)]



import re
from ..css_abbreviation import parse as abbreviation, tokens, CSSValue, CSSProperty, FunctionCall
from ..config import Config
from ..list_utils import some, get_item
from .snippets import create_snippet, nest, CSSSnippetProperty, CSSSnippetRaw, CSSSnippetType
from .score import calculate_score
from .color import color
from .format import stringify
from .scope import CSSAbbreviationScope

gradient_name = 'lg'


def parse(abbr: str, config: Config):
    """
    Parses given Emmet abbreviation into a final abbreviation tree with all
    required transformations applied
    """
    snippets = config.cache.get('stylesheet_snippets') if config.cache is not None else None

    if snippets is None:
        snippets = convert_snippets(config.snippets)
        if config.cache is not None:
            config.cache['stylesheet_snippets'] = snippets

    if isinstance(abbr, str):
        abbr = abbreviation(abbr, { 'value': is_value_scope(config) })

    filtered_snippets = get_snippets_for_scope(snippets, config)

    for node in abbr:
        resolve_node(node, filtered_snippets, config)

    return abbr


def convert_snippets(snippets: dict):
    "Converts given raw snippets into internal snippets representation"
    result = [create_snippet(k, v) for (k, v) in snippets.items()]
    return nest(result)


def resolve_node(node: CSSProperty, snippets: list, config: Config):
    """
    Resolves given node: finds matched CSS snippets using fuzzy match and resolves
    keyword aliases from node value
    """
    if not resolve_gradient(node, config):
        score = config.options.get('stylesheet.fuzzySearchMinScore', 0)
        if is_value_scope(config):
            # Resolve as value of given CSS property
            prop_name = config.context.get('name', '')
            snippet = None
            for s in snippets:
                if isinstance(s, CSSSnippetProperty) and s.property == prop_name:
                    snippet = s
                    break
            resolve_value_keywords(node, config, snippet, score)
            node.snippet = snippet
        elif node.name:
            snippet = find_best_match(node.name, snippets, score, True)
            node.snippet = snippet

            if snippet:
                if isinstance(snippet, CSSSnippetProperty):
                    resolve_as_property(node, snippet, config)
                else:
                    resolve_as_snippet(node, snippet)

    if node.name or config.context:
        # Resolve numeric values for CSS properties only
        resolve_numeric_value(node, config)

    return node


def resolve_gradient(node: CSSProperty, config: Config):
    "Resolves CSS gradient shortcut from given property, if possible"
    global gradient_name
    gradient_fn = None
    css_val = node.value[0] if len(node.value) == 1 else None

    if css_val is not None and len(css_val.value) == 1:
        v = css_val.value[0]
        if isinstance(v, FunctionCall) and v.name == gradient_name:
            gradient_fn = v

    if gradient_fn or node.name == gradient_name:
        gradient_value = gradient_fn.arguments if gradient_fn else [CSSValue([tokens.Field('', 0)])]
        gradient_fn = FunctionCall('linear-gradient', gradient_value)

        if not config.context:
            node.name = 'background-image'
        node.value = [CSSValue([gradient_fn])]

        return True

    return False


def resolve_as_property(node: CSSProperty, snippet: CSSSnippetProperty, config: Config):
    "Resolves given parsed abbreviation node as CSS property"

    abbr = node.name

    # Check for unmatched part of abbreviation
    # For example, in `dib` abbreviation the matched part is `d` and `ib` should
    # be considered as inline value. If unmatched fragment exists, we should check
    # if it matches actual value of snippet. If either explicit value is specified
    # or unmatched fragment did not resolve to to a keyword, we should consider
    # matched snippet as invalid
    inline_value = get_unmatched_part(abbr, snippet.key)
    if inline_value:
        if node.value:
            # Already have value: unmatched part indicates matched snippet is invalid
            return node

        kw = resolve_keyword(inline_value, config, snippet)
        if not kw:
            return node

        node.value.append(CSSValue([kw]))

    node.name = snippet.property

    if node.value:
        # Replace keyword alias from current abbreviation node with matched keyword
        resolve_value_keywords(node, config, snippet)
    elif snippet.value:
        default_value = snippet.value[0]

        # https://github.com/emmetio/emmet/issues/558
        # We should auto-select inserted value only if there’s multiple value
        # choice
        if len(snippet.value) == 1 or some(has_field, default_value):
            node.value = default_value
        else:
            node.value = list(map(lambda n: wrap_with_field(n, config), default_value))

    return node


def resolve_value_keywords(node: CSSProperty, config: Config, snippet: CSSSnippetProperty=None, minScore: int=0):
    for css_val in node.value:
        value = []

        for token in css_val.value:
            if isinstance(token, tokens.Literal):
                value.append(resolve_keyword(token.value, config, snippet, minScore) or token)
            elif isinstance(token, FunctionCall):
                # For function calls, we should find matching function call
                # and merge arguments
                match = resolve_keyword(token.name, config, snippet, minScore)
                if match and isinstance(match, FunctionCall):
                    value.append(FunctionCall(match.name, token.arguments + match.arguments[len(token.arguments):]))
                else:
                    value.append(token)
            else:
                value.append(token)

        css_val.value = value

def resolve_as_snippet(node: CSSProperty, snippet: CSSSnippetRaw):
    # When resolving snippets, we have to do the following:
    # 1. Replace field placeholders with actual field tokens.
    # 2. If input values given, put them instead of fields
    offset = 0
    input_value = get_item(node.value, 0)
    output_value = []

    for m in re.finditer(r'\$\{(\d+)(:[^}]+)?\}', snippet.value):
        if offset != m.start():
            output_value.append(tokens.Literal(snippet.value[offset:m.start()]))
        offset = m.end()
        if input_value and input_value.value:
            output_value.append(input_value.value.pop(0))
        else:
            output_value.append(tokens.Field(m.group(2)[1:] if m.group(2) else '', int(m.group(1))))

    tail = snippet.value[offset:]
    if tail:
        output_value.append(tokens.Literal(tail))

    node.name = None
    node.value = [CSSValue(output_value)]
    return node


def find_best_match(abbr: str, items: list, min_score=0, partial_match=False):
    """
    Finds best matching item from `items` array
    :param abbr  Abbreviation to match
    :param items List of items for match
    :param minScore The minimum score the best matched item should have to be a valid match.
    """
    max_score = 0
    matched_item = None

    for item in items:
        score = calculate_score(abbr, get_scoring_part(item), partial_match)

        if score == 1:
            # direct hit, no need to look further
            return item

        if score and score >= max_score:
            max_score = score
            matched_item = item

    return matched_item if max_score >= min_score else None


def get_scoring_part(item):
    return item if isinstance(item, str) else item.key


def get_unmatched_part(abbr: str, text: str):
    """
    Returns a part of `abbr` that wasn’t directly matched against `str`.
    For example, if abbreviation `poas` is matched against `position`,
    the unmatched part will be `as` since `a` wasn’t found in string stream
    """
    last_pos = 0
    for i, ch in enumerate(abbr):
        last_pos = text.find(ch, last_pos)
        if last_pos == -1:
            return abbr[i:]
        last_pos += 1

    return ''


def resolve_keyword(kw: str, config: Config, snippet: CSSSnippetProperty=None, min_score=0):
    """
    Resolves given keyword shorthand into matched snippet keyword or global keyword,
    if possible
    """
    if snippet:
        ref = find_best_match(kw, snippet.keywords.keys(), min_score)
        if ref:
            return snippet.keywords[ref]

        for dep in snippet.dependencies:
            ref = find_best_match(kw, dep.keywords.keys(), min_score)
            if ref:
                return dep.keywords[ref]

    ref = find_best_match(kw, config.options.get('stylesheet.keywords', []), min_score)
    if ref:
        return tokens.Literal(ref)

    return None


def resolve_numeric_value(node: CSSProperty, config: Config):
    "Resolves numeric values in given abbreviation node"
    aliases = config.options.get('stylesheet.unitAliases', {})
    unitless = config.options.get('stylesheet.unitless', [])

    for v in node.value:
        for t in v.value:
            if isinstance(t, tokens.NumberValue):
                if t.unit:
                    t.unit = aliases.get(t.unit, t.unit)
                elif t.value != 0 and node.name not in unitless:
                    opt_name = 'stylesheet.floatUnit' if '.' in t.raw_value else 'stylesheet.intUnit'
                    t.unit = config.options.get(opt_name, '')


def has_field(value: CSSValue):
    "Check if given value contains fields"
    for v in value.value:
        if isinstance(v, tokens.Field) or (isinstance(v, FunctionCall) and some(has_field, v.arguments)):
            return True

    return False


class WrapState:
    __slots__ = ('index',)

    def __init__(self, index=1):
        self.index = index

    def inc(self):
        index = self.index
        self.index += 1
        return index

def wrap_with_field(node: CSSValue, config: Config, state: WrapState=None):
    if state is None: state = WrapState()
    value = []

    for v in node.value:
        if isinstance(v, tokens.ColorValue):
            color_val = color(v, config.options.get('stylesheet.shortHex'))
            value.append(tokens.Field(color_val, state.inc()))
        elif isinstance(v, tokens.Literal):
            value.append(tokens.Field(v.value, state.inc()))
        elif isinstance(v, tokens.NumberValue):
            value.append(tokens.Field(''.join((v.value, v.unit)), state.inc()))
        elif isinstance(v, tokens.StringValue):
            q = '\'' if v.quote == 'single' else '"'
            value.append(tokens.Field(''.join((q, v.value, q)), state.inc()))
        elif isinstance(v, FunctionCall):
            value.append(tokens.Field(v.name, state.inc()))
            value.append(tokens.Literal('('))

            max_i = len(v.arguments) - 1
            for i, arg in enumerate(v.arguments):
                value += wrap_with_field(arg, config, state).value
                if i != max_i:
                    value.append(tokens.Literal(', '))

            value.append(tokens.Literal(')'))
        else:
            value.append(v)

    return CSSValue(value)

def is_value_scope(config: Config):
    "Check if abbreviation should be expanded in CSS value context"
    if config.context:
        return config.context['name'] == CSSAbbreviationScope.Value or not config.context['name'].startswith('@@')

    return False


def get_snippets_for_scope(snippets: list, config: Config):
    "Returns snippets for given scope"

    if config.context:
        if config.context['name'] == CSSAbbreviationScope.Section:
            return [s for s in snippets if s.type == CSSSnippetType.Raw]


        if config.context['name'] == CSSAbbreviationScope.Property:
            return [s for s in snippets if s.type == CSSSnippetType.Property]

    return snippets



from . import tokens
from .tokens import Chars, OperatorType
from ...scanner import Scanner
from ...scanner_utils import is_number, is_alpha, is_alpha_word, is_alpha_numeric_word, is_quote, is_space

OPERATOR_MAP = dict([
    (Chars.Sibling, OperatorType.Sibling),
    (Chars.Excl, OperatorType.Important),
    (Chars.Comma, OperatorType.ArgumentDelimiter),
    (Chars.Colon, OperatorType.PropertyDelimiter),
    (Chars.Dash, OperatorType.ValueDelimiter)
])

def tokenize(abbr: str, is_value=False):
    brackets = 0
    scanner = Scanner(abbr)
    result = []

    while not scanner.eof():
        token = field(scanner) or \
            number_value(scanner) or \
            color_value(scanner) or \
            string_value(scanner) or \
            bracket(scanner) or \
            operator(scanner) or \
            white_space(scanner) or \
            literal(scanner, brackets == 0 and not is_value) or \
            None

        if not token:
            raise scanner.error('Unexpected character')

        if isinstance(token, tokens.Bracket):
            if not brackets and token.open:
                merge_tokens(scanner, result)

            brackets += 1 if token.open else -1
            if brackets < 0:
                raise scanner.error('Unexpected bracket', token.start)

        result.append(token)

        # Forcibly consume next operator after unit-less numeric value or color:
        # next dash `-` must be used as value delimiter
        if should_consume_dash_after(token):
            token = operator(scanner)
            if token: result.append(token)

    return result

def field(scanner: Scanner):
    start = scanner.pos
    if scanner.eat(Chars.Dollar) and scanner.eat(Chars.CurlyBracketOpen):
        scanner.start = scanner.pos
        name = ''
        index = None

        if scanner.eat_while(is_number):
            # It’s a field
            index = int(scanner.current())
            name = consume_placeholder(scanner) if scanner.eat(Chars.Colon) else ''
        elif is_alpha(scanner.peek()):
            # It’s a variable
            name = consume_placeholder(scanner)

        if scanner.eat(Chars.CurlyBracketClose):
            return tokens.Field(name, index, start, scanner.pos)

        raise scanner.error('Expecting }')

    # If we reached here then there’s no valid field here, revert
    # back to starting position
    scanner.pos = start

def consume_placeholder(stream: Scanner):
    "Consumes a placeholder: value right after `:` in field. Could be empty"
    stack = []
    stream.start = stream.pos

    while not stream.eof():
        if stream.eat(Chars.CurlyBracketOpen):
            stack.append(stream.pos)
        elif stream.eat(Chars.CurlyBracketClose):
            if not stack:
                stream.pos -= 1
                break
            stack.pop()
        else:
            stream.pos += 1

    if stack:
        stream.pos = stack.pop()
        raise stream.error('Expecting }')

    return stream.current()

def literal(scanner: Scanner, short=False):
    """
    Consumes literal from given scanner
    :param short Use short notation for consuming value.
    The difference between “short” and “full” notation is that first one uses
    alpha characters only and used for extracting keywords from abbreviation,
    while “full” notation also supports numbers and dashes
    """
    start = scanner.pos

    if scanner.eat(is_ident_prefix):
        # SCSS or LESS variable
        # NB a bit dirty hack: if abbreviation starts with identifier prefix,
        # consume alpha characters only to allow embedded variables
        scanner.eat_while(is_keyword if start else is_literal)
    elif scanner.eat(is_alpha_word):
        scanner.eat_while(is_literal if short else is_keyword)
    else:
        # Allow dots only at the beginning of literal
        scanner.eat(Chars.Dot)
        scanner.eat_while(is_literal)

    if start != scanner.pos:
        scanner.start = start
        return create_literal(scanner)

def create_literal(scanner: Scanner, start=None, end=None):
    if start is None: start = scanner.start
    if end is None: end = scanner.pos
    return tokens.Literal(scanner.substring(start, end), start, end)

def number_value(scanner: Scanner):
    """
    Consumes numeric CSS value (number with optional unit) from current stream,
    if possible
    """
    start = scanner.pos
    if consume_number(scanner):
        scanner.start = start
        raw_value = scanner.current()
        value = float(raw_value)

        # eat unit, which can be a % or alpha word
        scanner.start = scanner.pos
        scanner.eat(Chars.Percent) or scanner.eat_while(is_alpha_word)
        return tokens.NumberValue(value, raw_value, scanner.current(), start, scanner.pos)

def string_value(scanner: Scanner):
    "Consumes quoted string value from given scanner"
    ch = scanner.peek()
    start = scanner.pos
    finished = False

    if is_quote(ch):
        scanner.pos += 1
        while not scanner.eof():
            # Do not throw error on malformed string
            if scanner.eat(ch):
                finished = True
                break
            else:
                scanner.pos += 1

        scanner.start = start
        value_start = start + 1
        value_end = scanner.pos - (1 if finished else 0)
        return tokens.StringValue(
            scanner.substring(value_start, value_end),
            'single' if ch == Chars.SingleQuote else 'double',
            start, scanner.pos)

def color_value(scanner: Scanner):
    "Consumes a color token from given scanner"
    # supported color variations:
    # #abc   → #aabbccc
    # #0     → #000000
    # #fff.5 → rgba(255, 255, 255, 0.5)
    # #t     → transparent

    start = scanner.pos
    if scanner.eat(Chars.Hash):
        value_start = scanner.pos
        color = ''
        alpha = ''
        if scanner.eat_while(is_hex):
            color = scanner.substring(value_start, scanner.pos)
            alpha = color_alpha(scanner)
        elif scanner.eat(Chars.Transparent):
            color = '0'
            alpha = color_alpha(scanner) or '0'
        else:
            alpha = color_alpha(scanner)

        if color or alpha or scanner.eof():
            r, g, b, a = parse_color(color, alpha)
            return tokens.ColorValue(r, g, b, a, scanner.substring(start + 1, scanner.pos), start, scanner.pos)
        else:
            # Consumed `#` but no actual value: invalid color value, treat it as literal
            return create_literal(scanner, start)

    scanner.pos = start

def color_alpha(scanner: Scanner):
    "Consumes alpha value of color: `.1`"
    start = scanner.pos
    if scanner.eat(Chars.Dot):
        scanner.start = start
        if scanner.eat_while(is_number):
            return scanner.current()
        return '1'

    return ''

def white_space(scanner: Scanner):
    "Consumes white space characters as string literal from given scanner"
    start = scanner.pos
    if scanner.eat_while(is_space):
        return tokens.WhiteSpace(start, scanner.pos)

def bracket(scanner: Scanner):
    "Consumes bracket from given scanner"
    ch = scanner.peek()
    if is_bracket(ch):
        start = scanner.pos
        scanner.pos += 1
        return tokens.Bracket(ch == Chars.RoundBracketOpen, start, scanner.pos)

def operator(scanner: Scanner):
    global OPERATOR_MAP
    op = OPERATOR_MAP.get(scanner.peek())
    if op:
        start = scanner.pos
        scanner.pos += 1
        return tokens.Operator(op, start, scanner.pos)

def consume_number(stream: Scanner):
    """
    Eats number value from given stream.
    Returns `True` if number was consumed
    """
    start = stream.pos
    stream.eat(Chars.Dash)
    after_negative = stream.pos

    has_decimal = stream.eat_while(is_number)

    prev_pos = stream.pos
    if stream.eat(Chars.Dot):
        # It’s perfectly valid to have numbers like `1.`, which enforces
        # value to float unit type
        has_float = stream.eat_while(is_number)
        if not has_decimal and not has_float:
            # Lone dot
            stream.pos = prev_pos

    # Edge case: consumed dash only: not a number, bail-out
    if stream.pos == after_negative:
        stream.pos = start

    return stream.pos != start

def is_ident_prefix(ch: str):
    return ch == Chars.At or ch == Chars.Dollar

def is_hex(ch: str):
    "Check if given code is a hex value (/0-9a-f/)"
    return is_number(ch) or 'A' <= ch <= 'F' or 'a' <= ch <= 'f'

def is_keyword(ch: str):
    return is_alpha_numeric_word(ch) or ch == Chars.Dash

def is_bracket(ch: str):
    return ch in (Chars.RoundBracketOpen, Chars.RoundBracketClose)

def is_literal(ch: str):
    return is_alpha_word(ch) or ch == Chars.Percent

def parse_color(value: str, alpha=None):
    r = '0'
    g = '0'
    b = '0'
    a = float(alpha) if alpha is not None and alpha != '' else 1

    if value == 't':
        a = 0
    else:
        l = len(value)
        if l == 0:
            pass
        elif l == 1:
            r = g = b = value + value
        elif l == 2:
            r = g = b = value
        elif l == 3:
            r = value[0] * 2
            g = value[1] * 2
            b = value[2] * 2
        else:
            value = value.rjust(6, '0')
            r = value[0:2]
            g = value[2:4]
            b = value[4:6]

    return (int(r, 16), int(g, 16), int(b, 16), a)


def should_consume_dash_after(token: tokens.Token):
    """
    Check if scanner reader must consume dash after given token.
    Used in cases where user must explicitly separate numeric values
    """
    return isinstance(token, tokens.ColorValue) or (isinstance(token, tokens.NumberValue) and not token.unit)


def merge_tokens(scanner: Scanner, token_list: list):
    """
    Merges last adjacent tokens into a single literal.
    This function is used to overcome edge case when function name was parsed
    as a list of separate tokens. For example, a `scale3d()` value will be
    parsed as literal and number tokens (`scale` and `3d`) which is a perfectly
    valid abbreviation but undesired result. This function will detect last adjacent
    literal and number values and combine them into single literal
    """
    start = 0
    end = 0

    while token_list:
        token = token_list[-1]
        if isinstance(token, tokens.Literal) or isinstance(token, tokens.NumberValue):
            start = token.start
            if not end: end = token.end
            token_list.pop()
        else:
            break

    if start != end:
        token_list.append(create_literal(scanner, start, end))



from .config import Config
from .markup import parse as markup_abbreviation, \
    stringify as stringify_markup, \
    abbreviation as parse_markup_abbreviation, \
    Abbreviation, AbbreviationNode, AbbreviationAttribute
from .stylesheet import parse as stylesheet_abbreviation, \
    stringify as stringify_stylesheet, \
    abbreviation as parse_stylesheet_abbreviation, \
    convert_snippets as parse_stylesheet_snippets
from .extract_abbreviation import extract_abbreviation as extract
from .scanner import ScannerException


def expand(abbr: str, config: dict={}, global_config: dict={}) -> str:
    "Expands given abbreviation into code snippet"
    if isinstance(config, Config):
        resolved_config = config
    else:
        resolved_config = Config(config, global_config)
    if resolved_config.type == 'stylesheet':
        return expand_stylesheet(abbr, resolved_config)

    return expand_markup(abbr, resolved_config)


def expand_markup(abbr: str, config: Config) -> str:
    """
    Expands given *markup* abbreviation (e.g. regular Emmet abbreviation that
    produces structured output like HTML) and outputs it according to options
    provided in config
    """
    return stringify_markup(markup_abbreviation(abbr, config), config)


def expand_stylesheet(abbr: str, config: Config):
    """
    Expands given *stylesheet* abbreviation (a special Emmet abbreviation designed for
    stylesheet languages like CSS, SASS etc.) and outputs it according to options
    provided in config
    """
    return stringify_stylesheet(stylesheet_abbreviation(abbr, config), config)



from ..abbreviation import parse, Abbreviation, AbbreviationNode, AbbreviationAttribute
from ..config import Config
from .utils import walk, find_deepest

def resolve_snippets(abbr: Abbreviation, config: Config):
    """
    Finds matching snippet from `registry` and resolves it into a parsed abbreviation.
    Resolved node is then updated or replaced with matched abbreviation tree.

    A HTML registry basically contains aliases to another Emmet abbreviations,
    e.g. a predefined set of name, attributes and so on, possibly a complex
    abbreviation with multiple elements. So we have to get snippet, parse it
    and recursively resolve it.
    """
    stack = []
    is_reversed = config.options.get('output.reverseAttributes', False)

    def resolve(child: AbbreviationNode):
        snippet = config.snippets.get(child.name) if child.name else None
        # A snippet in stack means circular reference.
        # It can be either a user error or a perfectly valid snippet like
        # "img": "img[src alt]/", e.g. an element with predefined shape.
        # In any case, simply stop parsing and keep element as is
        if not snippet or snippet in stack:
            return None

        snippet_abbr = parse(snippet, config)
        stack.append(snippet)
        walk_resolve(snippet_abbr, resolve, config)
        stack.pop()

        # Add attributes from current node into every top-level node of parsed abbreviation
        for top_node in snippet_abbr.children:
            if child.attributes:
                from_attr = top_node.attributes or []
                to_attr = child.attributes or []
                if is_reversed:
                    top_node.attributes = to_attr + from_attr
                else:
                    top_node.attributes = from_attr + to_attr
            merge(child, top_node)

        return snippet_abbr

    walk_resolve(abbr, resolve, config)
    return abbr


def walk_resolve(node: AbbreviationNode, resolve: callable, config: Config) -> list:
    children = []

    for child in node.children:
        resolved = resolve(child)

        if resolved:
            children += resolved.children

            deepest = find_deepest(resolved)
            if isinstance(deepest[1], AbbreviationNode):
                deepest[1].children += walk_resolve(child, resolve, config)
        else:
            children.append(child)
            child.children = walk_resolve(child, resolve, config)

    node.children = children
    return children


def merge(from_node: AbbreviationNode, to_node: AbbreviationNode):
    if from_node.self_closing:
        to_node.self_closing = True

    if from_node.value is not None:
        to_node.value = from_node.value

    if from_node.repeat:
        to_node.repeat = from_node.repeat



from .config import Config
from .abbreviation.convert import AbbreviationAttribute, AbbreviationNode

class OutputStream:
    __slots__ = ('options', '_value', 'level', 'offset', 'line', 'column')

    def __init__(self, options={}, level=0):
        self._value = []
        self.options = options
        self.level = level
        self.offset = 0
        self.line = 0
        self.column = 0

    @property
    def value(self):
        return ''.join(self._value)

    def _push(self, text: str):
        "Pushes raw string into output stream without any processing"
        l = len(text)
        self._value.append(text)
        self.offset += l
        self.column += l

    def push(self, text: str):
        "Pushes plain string into output stream without newline processing"
        process_text = self.options.get('output.text')
        self._push(process_text(text, offset=self.offset, line=self.line, column=self.column))

    def push_string(self, value: str):
        "Pushes given string with possible newline formatting into output"
        # If given value contains newlines, we should push content line-by-line and
        # use `push_newline()` to maintain proper line/column state
        first = True

        for line in value.splitlines():
            if not first: self.push_newline(True)
            first = False
            self.push(line)

    def push_newline(self, indent=None):
        "Pushes new line into given output stream"
        base_indent = self.options.get('output.baseIndent')
        newline = self.options.get('output.newline')
        self.push('%s%s' % (newline, base_indent))
        self.line += 1
        self.column = len(base_indent)
        if indent:
            self.push_indent(self.level if indent is True else indent)

    def push_indent(self, size=None):
        "Adds indentation of `size` to current output stream"

        if size is None: size = self.level
        indent = self.options.get('output.indent')
        self.push(indent * max(size, 0))

    def push_field(self, index: int, placeholder: str=''):
        field = self.options.get('output.field')
        # NB: use `_push` instead of `push` to skip text processing
        self._push(field(index, placeholder, offset=self.offset, line=self.line, column=self.column))


def tag_name(name: str, config: Config):
    "Returns given tag name formatted according to given config"
    return str_case(name, config.options.get('output.tagCase'))


def attr_name(name: str, config: Config):
    "Returns given attribute name formatted according to given config"
    return str_case(name, config.options.get('output.attributeCase'))


def attr_quote(attr: AbbreviationAttribute, config: Config, is_open: bool=None):
    "Returns character for quoting value of given attribute"
    if attr.value_type == 'expression':
        return '{' if is_open else '}'

    return '\'' if config.options.get('output.attributeQuotes') == 'single' else '"'


def is_boolean_attribute(attr: AbbreviationAttribute, config: Config):
    "Check if given attribute is boolean"
    if attr.boolean:
        return True

    name = (attr.name or '').lower()
    return name in config.options.get('output.booleanAttributes', [])

def self_close(config: Config):
    "Returns a token for self-closing tag, depending on current options"
    style = config.options.get('output.selfClosingStyle')
    if style == 'xhtml': return ' /'
    if style == 'xml': return '/'
    return ''

def is_inline(node: AbbreviationNode, config: Config):
    if isinstance(node, str):
        return node.lower() in config.options.get('inlineElements', [])

    # inline node is a node either with inline-level name or text-only node
    return is_inline(node.name, config) if node.name else bool(node.value and not node.attributes)

def str_case(text: str, case_type: str):
    if case_type:
        return text.upper() if case_type == 'upper' else text.lower()

    return text




vocabulary = {
	"common": ["mujer", "uno", "dolor", "más", "de", "poder", "mismo", "si"],
	"words": ["ejercicio", "preferencia", "perspicacia", "laboral", "paño",
		"suntuoso", "molde", "namibia", "planeador", "mirar", "demás", "oficinista", "excepción",
		"odio", "consecuencia", "casi", "auto", "chicharra", "velo", "elixir",
		"ataque", "no", "odio", "temporal", "cuórum", "dignísimo",
		"facilismo", "letra", "nihilista", "expedición", "alma", "alveolar", "aparte",
		"león", "animal", "como", "paria", "belleza", "modo", "natividad",
		"justo", "ataque", "séquito", "pillo", "sed", "ex", "y", "voluminoso",
		"temporalidad", "verdades", "racional", "asunción", "incidente", "marejada",
		"placenta", "amanecer", "fuga", "previsor", "presentación", "lejos",
		"necesariamente", "sospechoso", "adiposidad", "quindío", "pócima",
		"voluble", "débito", "sintió", "accesorio", "falda", "sapiencia",
		"volutas", "queso", "permacultura", "laudo", "soluciones", "entero",
		"pan", "litro", "tonelada", "culpa", "libertario", "mosca", "dictado",
		"reincidente", "nascimiento", "dolor", "escolar", "impedimento", "mínima",
		"mayores", "repugnante", "dulce", "obcecado", "montaña", "enigma",
		"total", "deletéreo", "décima", "cábala", "fotografía", "dolores",
		"molesto", "olvido", "paciencia", "resiliencia", "voluntad", "molestias",
		"magnífico", "distinción", "ovni", "marejada", "cerro", "torre", "y",
		"abogada", "manantial", "corporal", "agua", "crepúsculo", "ataque", "desierto",
		"laboriosamente", "angustia", "afortunado", "alma", "encefalograma",
		"materialidad", "cosas", "o", "renuncia", "error", "menos", "conejo",
		"abadía", "analfabeto", "remo", "fugacidad", "oficio", "en", "almácigo", "vos", "pan",
		"represión", "números", "triste", "refugiado", "trote", "inventor",
		"corchea", "repelente", "magma", "recusado", "patrón", "explícito",
		"paloma", "síndrome", "inmune", "autoinmune", "comodidad",
		"ley", "vietnamita", "demonio", "tasmania", "repeler", "apéndice",
		"arquitecto", "columna", "yugo", "computador", "mula", "a", "propósito",
		"fantasía", "alias", "rayo", "tenedor", "deleznable", "ventana", "cara",
		"anemia", "corrupto"]
}
